from ..headers._bsl_inst_info import _bsl_inst_info_list as inst
from ..headers._bsl_logger import _bsl_logger as bsl_logger
from ..headers._bsl_type import _bsl_type as bsl_type
import numpy
try:
    import seabreeze.spectrometers as sb
except Exception:
    sb = None
from numpy.typing import NDArray
import time

class HR4000CG:
    CONNECT_RETRY_COUNT = 3
    CONNECT_RETRY_DELAY_SEC = 0.5

    def __init__(self, device_sn:str=None) -> None:
        """
        Initialize and connect an HR4000CG spectrometer.

        Parameters
        ----------
        device_sn : str | None, optional
            Optional serial selector, by default None.
        """
        self.inst = inst.HR4000CG
        self.target_device_sn = device_sn
        self.device_id=""
        self.device_model=""
        self.logger = bsl_logger(self.inst)
        self.logger.info(f"Initiating bsl_instrument - SPEC({device_sn})...")
        
        self.__connect_spectrometer()
            
        if self.spec is not None:
            self.logger.device_id = self.device_id
            self.logger.success(f"READY - OceanOptics HR4000CG Spectrometer \"{self.device_id}\"\n\n")
        return None

    def __del__(self, *args, **kwargs) -> None:
        try:
            self.close()
        except Exception:
            pass
        return None

    def __connect_spectrometer(self) -> None:
        """
        - Try to connect to spectrometer based on the s/n provided, if no
        s/n is provided, connect to the next available spectrometer on the bus.

        Raises:
            self.DeviceConnectionFailed: Failed to connect to spectrometer.
        """
        if sb is None:
            self.logger.error("seabreeze is not installed; HR4000CG communication unavailable.")
            raise bsl_type.DeviceConnectionFailed
        self.spec = None
        devices = sb.list_devices()
        if len(devices) == 0:
            self.logger.error(f"Device not found on communication bus.\n\n\n")
            raise bsl_type.DeviceConnectionFailed
        
        self.logger.trace(f"Devices found on bus: {str(devices)}")
        last_error = None
        for attempt in range(1, self.CONNECT_RETRY_COUNT + 1):
            try:
                if self.target_device_sn in (None, ""):
                    self.spec = sb.Spectrometer.from_first_available()
                elif self.target_device_sn in str(devices):
                    self.spec = sb.Spectrometer.from_serial_number(self.target_device_sn)
                else:
                    self.logger.error(
                        f"FAILED - Device[s] found on the bus, but failed to find requested device with s/n: \"{self.target_device_sn}\".\n\n\n"
                    )
                    raise bsl_type.DeviceConnectionFailed
                break
            except Exception as exc:
                last_error = exc
                self.spec = None
                self.logger.warning(
                    f"Connect attempt {attempt}/{self.CONNECT_RETRY_COUNT} failed for HR4000CG."
                )
                time.sleep(self.CONNECT_RETRY_DELAY_SEC)

        if self.spec is None:
            self.logger.error(
                f"FAILED - Device[s] found on the communication bus, but failed to make connection.\n\n\n"
            )
            raise bsl_type.DeviceConnectionFailed from last_error
            
        self.device_id = self.spec.serial_number
        self.device_model = self.spec.model
        return None

    def reconnect(self, device_sn: str | None = None) -> bool:
        """
        Reconnect to spectrometer.

        Parameters
        ----------
        device_sn : str | None, optional
            Optional serial selector override, by default None.

        Returns
        -------
        bool
            True when reconnection succeeds.
        """
        if device_sn is not None:
            self.target_device_sn = device_sn
        self.close()
        try:
            self.__connect_spectrometer()
            return self.spec is not None
        except Exception:
            return False

    def reset_connection(self) -> bool:
        """
        Reset spectrometer connection using current selector.

        Returns
        -------
        bool
            True when reconnection succeeds.
        """
        return self.reconnect(device_sn=self.target_device_sn)

    def get_wavelength(self) -> NDArray[numpy.float64]:
        """
        - wavelength array of the spectrometer
        - wavelengths in (nm) corresponding to each pixel of the spectrometer

        Returns
        -------
        wavelengths : `numpy.ndarray`
            wavelengths in (nm)
        """
        return self.spec.wavelengths()

    def get_intensity(self, correct_dark_counts: bool = False, correct_nonlinearity: bool = False) -> NDArray[numpy.float64]:
        """
        - measured intensity array in (a.u.)

        Measured intensities as numpy array returned by the spectrometer.
        The measuring behavior can be adjusted by setting the trigger mode.
        Pixels at the start and end of the array might not be optically
        active so interpret their returned measurements with care. Refer
        to the spectrometer's datasheet for further information.

        Parameters
        ----------
        correct_dark_counts : `bool`
            If requested and supported the average value of electric dark
            pixels on the ccd of the spectrometer is subtracted from the
            measurements to remove the noise floor in the measurements
            caused by non optical noise sources.
        correct_nonlinearity : `bool`
            Some spectrometers store non linearity correction coefficients
            in their eeprom. If requested and supported by the spectrometer
            the readings returned by the spectrometer will be linearized
            using the stored coefficients.

        Returns
        -------
        intensities : `numpy.ndarray`
            measured intensities in (a.u.)
        """
        return self.spec.intensities(correct_dark_counts, correct_nonlinearity)

    def get_spectrum(self, correct_dark_counts:bool=False, correct_nonlinearity:bool=False) -> NDArray[numpy.float64]:
        """
        - returns wavelengths and intensities as single array

        Uses
        ----------
        >>> (wavelengths, intensities) = spec.spectrum()

        Parameters
        ----------
        correct_dark_counts : `bool`
            see `Spectrometer.intensities`
        correct_nonlinearity : `bool`
            see `Spectrometer.intensities`

        Returns
        -------
        spectrum : `numpy.ndarray`
            combined array of wavelengths and measured intensities
        """
        temp = self.spec.spectrum(correct_dark_counts, correct_nonlinearity)
        return self.spec.spectrum(correct_dark_counts, correct_nonlinearity)

    def set_integration_time_micros(self, exp_us:int) -> None:
        """
        - set the integration time in microseconds

        Parameters
        ----------
        integration_time_micros : `int`
            integration time in microseconds
        """
        self.spec.integration_time_micros(exp_us)
        return None

    @property
    def integration_time_limit_us(self) -> 'tuple[int,int]':
        """
        - return the hardcoded minimum and maximum integration time

        Returns
        -------
        integration_time_micros_min_max : `tuple[int, int]`
            min and max integration time in micro seconds
        """
        return self.spec.integration_time_micros_limits
    
    @property
    def device_max_intensity(self) -> float:
        """
        - return the maximum intensity of the spectrometer

        Returns
        -------
        max_intensity : `float`
            the maximum intensity that can be returned by the spectrometer in (a.u.)
            It's possible that the spectrometer saturates already at lower values.
        """
        return self.spec.max_intensity

    @property
    def device_pixel_count(self) -> int:
        """the spectrometer's number of pixels"""
        return self.spec.pixels

    def close(self) -> None:
        """
        Close spectrometer resources safely.
        """
        try:
            spec_obj = getattr(self, "spec", None)
            if spec_obj is not None:
                spec_obj.close()
        except Exception:
            pass
        self.spec = None
        self.logger.success(f"CLOSED - OceanOptics HR4000CG Spectrometer \"{self.device_id}\"\n\n\n")
        return None



from ..interfaces._bsl_serial import _bsl_serial as bsl_serial
from ..headers._bsl_inst_info import _bsl_inst_info_list as inst
from ..headers._bsl_logger import _bsl_logger as bsl_logger
from ..headers._bsl_type import _bsl_type as bsl_type
import time, re, enum

class USB_520:
    CONNECT_RETRY_COUNT = 3
    CONNECT_RETRY_DELAY_SEC = 0.5

    class USB_520_SN(enum.Enum):
        CH1 = '1066656'; CH2 = '1066657'; CH3 = '1066658'; CH4 = '1066659'

    def __init__(self, device_sn='', tear_on_startup:bool = True, reverse_negative:bool = True) -> None:
        """
        Read the latest force measurement from the device in grams.

        Parameters
        ----------
        device_sn : `str`, optional
            The serial number of the device to connect to. 
            If not specified, the first available device will be connected.
        
        tear_on_startup : `bool`, optional
            If True, the tear calibration will be performed on startup.
            The default is True.
        
        reverse_negative : `bool`, optional
            If True, the force reading will be reversed for positive values.
            The default is True.
            
        Returns
        -------
        status : `int`
            0 if successful, otherwise raise an exception.
        """
        if "CH" in device_sn:
            try:
                device_sn = self.USB_520_SN[device_sn].value
            except KeyError as exc:
                raise bsl_type.DeviceConnectionFailed(
                    f"Unknown USB_520 channel alias: {device_sn}"
                ) from exc
            
        self._target_device_sn = device_sn
        self.inst = inst.USB_520
        self.flip_result = reverse_negative
        self.device_id = ""
        self.logger = bsl_logger(self.inst)
        self.logger.info(f"Initiating bsl_instrument - Futek USB_520({device_sn})...")

        self.serial = None
        if self._target_device_sn != "":
            if self._serial_connect():
                self.logger.device_id = self.serial.device_id.split('-')[-1]
                self.__system_init(tear_on_startup)
                self.logger.success(f"READY - Futek USB DAC.\n\n\n")
                return None
            self.logger.error(f"FAILED to connect to Futek USB DAC!\n\n\n")
            raise bsl_type.DeviceConnectionFailed

        for target_sn in self.USB_520_SN:
            target_sn = target_sn.value
            self._target_device_sn = target_sn
            if self._serial_connect():
                self.logger.device_id = self.serial.device_id.split('-')[-1]
                self.__system_init(tear_on_startup)
                self.logger.success(f"READY - Futek USB DAC.\n\n\n")
                return None
        
        self.logger.error(f"FAILED to connect to Futek USB DAC!\n\n\n")
        raise bsl_type.DeviceConnectionFailed

    def __del__(self, *args, **kwargs) -> None:
        try:
            self.close()
        except Exception:
            pass
        return None
    
    
    def __system_init(self, tear_on_startup:bool = True) -> int:
        self.tear_calibration = 0.0
        if tear_on_startup:
            self.set_tear_calibration()
        return 0
    

    def _serial_connect(self) -> bool:
        """
        Connect to USB-520 serial interface with bounded retries.

        Returns
        -------
        bool
            True when connection succeeds.
        """
        self.serial = None
        last_error = None
        for attempt in range(1, self.CONNECT_RETRY_COUNT + 1):
            try:
                candidate = bsl_serial(inst.USB_520, self._target_device_sn)
                if candidate is None or getattr(candidate, "serial_port", None) is None:
                    raise bsl_type.DeviceConnectionFailed("No serial communication port found.")
                if not candidate.serial_port.is_open:
                    raise bsl_type.DeviceConnectionFailed("Serial port is not open.")
                self.serial = candidate
                return True
            except Exception as exc:
                last_error = exc
                self.logger.warning(
                    f"Connection attempt {attempt}/{self.CONNECT_RETRY_COUNT} failed: {type(exc)}"
                )
                time.sleep(self.CONNECT_RETRY_DELAY_SEC)
        self.logger.error(f"Unable to connect USB_520 after retries: {repr(last_error)}")
        return False

    def reconnect(self, device_sn: str = "", tear_on_startup: bool = False) -> bool:
        """
        Reconnect to USB-520 device.

        Parameters
        ----------
        device_sn : str, optional
            Optional serial selector override, by default ``""``.
        tear_on_startup : bool, optional
            Run tare calibration after reconnect, by default False.

        Returns
        -------
        bool
            True when reconnection succeeds.
        """
        if device_sn:
            self._target_device_sn = device_sn
        self.close()
        if not self._serial_connect():
            return False
        self.__system_init(tear_on_startup=tear_on_startup)
        return True

    def reset_tear_calibration(self, average_count: int = 10) -> bool:
        """
        Re-run tare calibration from fresh measurements.

        Parameters
        ----------
        average_count : int, optional
            Number of samples for tare averaging, by default 10.

        Returns
        -------
        bool
            True when calibration succeeds.
        """
        try:
            self.set_tear_calibration(average_count=average_count)
            return True
        except Exception as exc:
            self.logger.error(f"Failed to reset USB_520 tare calibration: {type(exc)}")
            return False
    

    def __extract_float(self, msg:str) -> float:
        match = re.search(r'([+-]?\d+\.\d+)\s*g', msg)
        if match:
            return float(match.group(1))
        else:
            self.logger.warning("Unable to find a force reading! please check the device.")
            return 999
    
    
    def get_new_measurement(self, timeout_ms:int = 10000, enable_tear:bool = True) -> float:
        """
        Read the latest force measurement from the device in grams.

        Parameters
        ----------
        timeout_ms : `int`, optional
            The timeout in milliseconds for the measurement to be read back from the device. 
            The default is 10000ms.

        enable_tear : `bool`, optional
            If True, the tear calibration value will be subtracted from the measurement.
            The default is True.
        
        Returns
        -------
        force : `float`
            The force value in grams read back from the sensor.
        """
        self.serial.serial_port.timeout = (timeout_ms/1000)
        self.serial.flush_read_buffer()
        msg = ""
        start = time.time()

        self.logger.debug("Waiting for new measurement...")
        while ("g" not in msg) and ((time.time()-start) < (timeout_ms/1000)):
            msg = self.serial.readline()
            if "g" in msg:
                self.logger.debug(f"New measurement received: {msg}")
                force = self.__extract_float(msg)
                if force == 999:
                    msg = ""
                    self.logger.warning("Retrying...")
                    pass
                else:
                    if self.flip_result:
                        force *= -1
                    if enable_tear:
                        force -= self.tear_calibration
                    return force

        self.logger.warning(f"Timeout! No new measurement received in {timeout_ms}ms.")
        raise bsl_type.DeviceTimeOutError
    
    
    def set_tear_calibration(self, average_count:int=10) -> float:
        """
        Read new force measurements and set it as the tear calibration value.
        
        Parameters
        ----------
        average_count : `int`, optional
            The number of measurements to average for the calibration.

        Returns
        -------
        force : `float`
            The force value in grams for the calibration.
        """
        self.logger.debug("Setting tear calibration...")
        count = 0
        sum = 0.0
        while (count < average_count):
            sum += self.get_new_measurement(enable_tear=False)
            count += 1
        self.tear_calibration = sum/count
        self.logger.success(f"Tear calibration set to {self.tear_calibration} grams.")
        return self.tear_calibration
    

    def close(self) -> None:
        """
        Close USB-520 communication resources safely.
        """
        serial_obj = getattr(self, "serial", None)
        if serial_obj is not None:
            try:
                serial_obj.close()
            except Exception:
                pass
            self.serial = None
        self.logger.success(f"CLOSED - Futek USB DAC.\n\n\n")
        pass



from __future__ import annotations

from functools import wraps
from typing import Any, TYPE_CHECKING, Callable, cast
import weakref

from loguru import logger

from ..core.device_health import device_health_hub
from ..core.device_monitor_gui import start_device_monitor_window
from ..core.instrument_runtime import InstrumentRecoveryManager, RecoveryPolicy, SafeInstrumentProxy
from ..core.logging import init_logger as _init_core_logger
from .factory import create_instrument

if TYPE_CHECKING:
    from ._inst_lib.instruments._BSC203 import BSC203_HDR50 as BSC203_HDR50Driver
    from ._inst_lib.instruments._CS260B import CS260B as CS260BDriver
    from ._inst_lib.instruments._DC2200 import DC2200 as DC2200Driver
    from ._inst_lib.instruments._Futek_USB_520 import USB_520 as USB520Driver
    from ._inst_lib.instruments._HR4000CG import HR4000CG as HR4000CGDriver
    from ._inst_lib.instruments._M69920 import M69920 as M69920Driver
    from ._inst_lib.instruments._PM100D import PM100D as PM100DDriver
    from ._inst_lib.instruments._PM400 import PM400 as PM400Driver
    from ._inst_lib.instruments._RS_7_1 import RS_7_1 as RS_7_1Driver
    from ._inst_lib.instruments._SP_2150 import SP_2150 as SP_2150Driver
    from ._inst_lib.instruments._mantisCam import MantisCamCtrl


__is_logger_ready = False
__GLOBAL_LOG_LEVEL = "DEBUG"
__is_monitor_started = False
_DEL_HOOKED_CLASSES: set[type] = set()


def init_logger(LOG_LEVEL: str = "DEBUG") -> None:
    """
    Initialize shared logging for instrument operations.

    Parameters
    ----------
    LOG_LEVEL : str, optional
        Log level string accepted by Loguru, by default ``"DEBUG"``.

    Returns
    -------
    None
    """
    global __is_logger_ready
    global __GLOBAL_LOG_LEVEL
    __GLOBAL_LOG_LEVEL = LOG_LEVEL
    _init_core_logger(log_level=LOG_LEVEL)
    __is_logger_ready = True


def _ensure_logger_ready() -> None:
    """
    Ensure logger initialization has occurred.

    Returns
    -------
    None
    """
    if not __is_logger_ready:
        init_logger()


def _ensure_monitor_window() -> None:
    """
    Launch the optional device monitor window once per process.

    Returns
    -------
    None
    """
    global __is_monitor_started
    if __is_monitor_started:
        return
    try:
        __is_monitor_started = start_device_monitor_window()
    except Exception as exc:
        logger.warning("Device monitor window startup failed: {}", exc)
        __is_monitor_started = False


def _extract_device_info(instrument_name: str, device: Any) -> tuple[str, str, str]:
    """
    Extract display metadata for monitoring.

    Parameters
    ----------
    instrument_name : str
        Requested logical instrument key.
    device : Any
        Constructed instrument object.

    Returns
    -------
    tuple[str, str, str]
        ``(model, device_type, serial_number)``.
    """
    inst_meta = getattr(device, "inst", None)
    model = str(getattr(inst_meta, "MODEL", instrument_name))
    device_type = str(getattr(inst_meta, "TYPE", "Unknown"))

    serial_number = (
        getattr(device, "device_id", "")
        or getattr(device, "device_sn", "")
        or getattr(device, "target_device_sn", "")
    )

    if instrument_name == "mantisCam":
        try:
            identity_cb = getattr(device, "get_camera_identity", None)
            identity = identity_cb(refresh=True) if callable(identity_cb) else {}
            if isinstance(identity, dict):
                nickname = str(identity.get("nickname") or identity.get("display_name") or "").strip()
                camera_model = str(identity.get("model") or identity.get("model-name") or "").strip()
                if nickname:
                    model = nickname
                elif camera_model:
                    model = camera_model

                serial_candidate = str(
                    identity.get("serial") or identity.get("serial-number") or serial_number or ""
                ).strip()
                if serial_candidate:
                    serial_number = serial_candidate
                    try:
                        device.device_id = serial_candidate
                    except Exception:
                        pass

                camera_type = str(
                    identity.get("camera_type") or identity.get("camera_type_str") or ""
                ).strip()
                device_type = f"MantisCam/{camera_type}" if camera_type else "MantisCam"
            else:
                device_type = "MantisCam"
        except Exception:
            device_type = "MantisCam"

    if serial_number is None or str(serial_number).strip() == "":
        serial_number = "Unknown"
    return str(model), str(device_type), str(serial_number)


def _set_attr_safely(device: Any, name: str, value: Any) -> bool:
    """
    Attempt to set an attribute on an instrument instance.

    Parameters
    ----------
    device : Any
        Target instrument instance.
    name : str
        Attribute name.
    value : Any
        Attribute value.

    Returns
    -------
    bool
        True when assignment succeeds.
    """
    try:
        setattr(device, name, value)
        return True
    except Exception:
        return False


def _ensure_class_del_hook(device: Any) -> bool:
    """
    Install a class-level ``__del__`` hook that auto-releases managed instances.

    Parameters
    ----------
    device : Any
        Instrument instance whose class may need hook installation.

    Returns
    -------
    bool
        True when hook is installed or already available.
    """
    cls = type(device)
    if cls in _DEL_HOOKED_CLASSES:
        return True

    original_del = getattr(cls, "__del__", None)
    if getattr(original_del, "__bsl_del_wrapped__", False):
        _DEL_HOOKED_CLASSES.add(cls)
        return True

    def _del_with_release(self: Any) -> None:
        release_error = ""
        runtime = getattr(self, "_bsl_runtime", None)
        if isinstance(runtime, dict):
            close_cb = getattr(self, "close", None)
            close_is_monitored = bool(callable(close_cb) and getattr(close_cb, "__bsl_close_wrapped__", False))
            try:
                if callable(close_cb):
                    close_cb()
                else:
                    raise RuntimeError("Close callback is not callable")
            except Exception as exc:
                release_error = str(exc)
                if not close_is_monitored:
                    try:
                        device_health_hub.register_disconnection(
                            instrument_key=str(runtime.get("instrument_name", "Unknown")),
                            model=str(runtime.get("model", "Unknown")),
                            device_type=str(runtime.get("device_type", "Unknown")),
                            serial_number=str(runtime.get("serial_number", "Unknown")),
                            error_message=release_error,
                        )
                    except Exception:
                        pass

        if callable(original_del):
            try:
                original_del(self)
            except Exception:
                pass

    _set_attr_safely(_del_with_release, "__bsl_del_wrapped__", True)
    try:
        setattr(cls, "__del__", _del_with_release)
        _DEL_HOOKED_CLASSES.add(cls)
        return True
    except Exception:
        return False


def _finalize_instrument_release(
    instrument_name: str,
    model: str,
    device_type: str,
    serial_number: str,
    close_state: dict[str, bool],
    device_ref: "weakref.ReferenceType[Any]",
) -> None:
    """
    Finalizer callback that closes an instrument when it is garbage-collected.

    Parameters
    ----------
    instrument_name : str
        Logical instrument key.
    model : str
        Device model at registration time.
    device_type : str
        Device type at registration time.
    serial_number : str
        Device serial at registration time.
    close_state : dict[str, bool]
        Mutable close state shared with close wrapper.
    device_ref : weakref.ReferenceType[Any]
        Weak reference to the instrument object.

    Returns
    -------
    None
    """
    if bool(close_state.get("closed", False)):
        return

    error_message = ""
    latest_model = model
    latest_type = device_type
    latest_serial = serial_number

    try:
        device = device_ref()
        if device is not None:
            try:
                latest_model, latest_type, latest_serial = _extract_device_info(instrument_name, device)
            except Exception:
                pass

            close_cb = getattr(device, "close", None)
            if callable(close_cb):
                close_is_monitored = bool(getattr(close_cb, "__bsl_close_wrapped__", False))
                try:
                    close_cb()
                except Exception as exc:
                    error_message = str(exc)
                # Monitored close already updates monitor state.
                if close_is_monitored:
                    return

    except Exception as exc:
        error_message = str(exc)

    try:
        device_health_hub.register_disconnection(
            instrument_key=instrument_name,
            model=latest_model or model,
            device_type=latest_type or device_type,
            serial_number=latest_serial or serial_number,
            error_message=error_message,
        )
    except Exception:
        pass


def _attach_runtime_hooks(
    *,
    instrument_name: str,
    device: Any,
    model: str,
    device_type: str,
    serial_number: str,
) -> None:
    """
    Attach additive runtime recovery helpers to a concrete instrument object.

    Parameters
    ----------
    instrument_name : str
        Logical instrument key.
    device : Any
        Instrument object returned by the driver constructor.
    model : str
        Display model name.
    device_type : str
        Display device type.
    serial_number : str
        Device serial number.

    Returns
    -------
    None

    Notes
    -----
    These hooks are additive and never replace normal driver methods except for
    an instance-level close wrapper used to publish monitor state transitions.
    """
    recovery = InstrumentRecoveryManager(
        device=device,
        instrument_name=instrument_name,
        policy=RecoveryPolicy(),
    )

    safe_name = "safe" if not hasattr(device, "safe") else "bsl_safe"
    _set_attr_safely(device, safe_name, SafeInstrumentProxy(recovery))

    invoke_target: Callable[..., Any] = recovery.invoke
    if not callable(getattr(device, "invoke", None)):
        _set_attr_safely(device, "invoke", invoke_target)
    else:
        _set_attr_safely(device, "invoke_safe", invoke_target)

    def _reconnect_safe() -> bool:
        ok = recovery.reconnect()
        if not ok:
            return False
        try:
            next_model, next_type, next_sn = _extract_device_info(instrument_name, device)
            device_health_hub.register_connection(
                instrument_key=instrument_name,
                model=next_model,
                device_type=next_type,
                serial_number=next_sn,
            )
        except Exception:
            pass
        return True

    _set_attr_safely(device, "reconnect_safe", _reconnect_safe)
    _set_attr_safely(device, "reset_safe", recovery.reset)
    _set_attr_safely(
        device,
        "_bsl_runtime",
        {
            "instrument_name": instrument_name,
            "model": model,
            "device_type": device_type,
            "serial_number": serial_number,
        },
    )

    if instrument_name == "mantisCam":
        monitor_setter = getattr(device, "set_monitor_callbacks", None)
        if callable(monitor_setter):
            def _on_mantis_connected(next_model: str, next_type: str, next_sn: str, _detail: str) -> None:
                try:
                    device_health_hub.register_connection(
                        instrument_key=instrument_name,
                        model=next_model or model,
                        device_type=next_type or device_type,
                        serial_number=next_sn or serial_number,
                    )
                except Exception:
                    pass

            def _on_mantis_warning(next_model: str, next_type: str, next_sn: str, detail: str) -> None:
                try:
                    device_health_hub.register_failure(
                        instrument_key=instrument_name,
                        model=next_model or model,
                        device_type=next_type or device_type,
                        serial_number=next_sn or serial_number,
                        error_message=detail or "MantisCam runtime warning.",
                        unrecoverable=False,
                    )
                except Exception:
                    pass

            def _on_mantis_disconnected(next_model: str, next_type: str, next_sn: str, detail: str) -> None:
                try:
                    device_health_hub.register_disconnection(
                        instrument_key=instrument_name,
                        model=next_model or model,
                        device_type=next_type or device_type,
                        serial_number=next_sn or serial_number,
                        error_message=detail or "MantisCam backend disconnected.",
                    )
                except Exception:
                    pass

            try:
                monitor_setter(
                    on_connected=_on_mantis_connected,
                    on_warning=_on_mantis_warning,
                    on_disconnected=_on_mantis_disconnected,
                )
            except Exception:
                pass

    del_hook_ready = _ensure_class_del_hook(device)

    close_cb = getattr(device, "close", None)
    close_state = {"closed": False}

    if callable(close_cb) and not getattr(close_cb, "__bsl_close_wrapped__", False):
        @wraps(close_cb)
        def _close_with_monitor(*args: Any, **kwargs: Any) -> Any:
            if close_state["closed"]:
                return None
            close_state["closed"] = True

            error_message = ""
            try:
                return close_cb(*args, **kwargs)
            except Exception as exc:
                error_message = str(exc)
                raise
            finally:
                try:
                    close_model, close_type, close_sn = _extract_device_info(instrument_name, device)
                    if close_model in {"", "Unknown", instrument_name}:
                        close_model = model
                    if close_type in {"", "Unknown"}:
                        close_type = device_type
                    if close_sn in {"", "Unknown"}:
                        close_sn = serial_number
                    device_health_hub.register_disconnection(
                        instrument_key=instrument_name,
                        model=close_model or model,
                        device_type=close_type or device_type,
                        serial_number=close_sn or serial_number,
                        error_message=error_message,
                    )
                except Exception:
                    pass

        _set_attr_safely(_close_with_monitor, "__bsl_close_wrapped__", True)
        _set_attr_safely(device, "close", _close_with_monitor)

    # If class-level __del__ hook is unavailable, keep a weakref finalizer fallback
    # so monitor state still transitions to disconnected.
    if not del_hook_ready:
        try:
            device_ref = weakref.ref(device)
            gc_finalizer = weakref.finalize(
                device,
                _finalize_instrument_release,
                instrument_name,
                model,
                device_type,
                serial_number,
                close_state,
                device_ref,
            )
            _set_attr_safely(device, "_bsl_gc_finalizer", gc_finalizer)
        except Exception:
            pass


def _build_instrument(instrument_name: str, *args: Any, **kwargs: Any) -> Any:
    """
    Build an instrument object with monitoring hooks.

    Parameters
    ----------
    instrument_name : str
        Logical instrument key.
    *args : Any
        Positional constructor arguments.
    **kwargs : Any
        Keyword constructor arguments.

    Returns
    -------
    Any
        Concrete instrument object from the requested driver.
    """
    _ensure_logger_ready()
    _ensure_monitor_window()

    serial_hint = str(kwargs.get("device_sn", "")).strip()
    if not serial_hint and args and isinstance(args[0], str):
        serial_hint = args[0].strip()

    connecting_type = "MantisCam" if instrument_name == "mantisCam" else instrument_name
    try:
        device_health_hub.register_connecting(
            instrument_key=instrument_name,
            model=instrument_name,
            device_type=connecting_type,
            serial_number=serial_hint or "Unknown",
            note="Connection in progress.",
        )
    except Exception:
        pass

    try:
        device = create_instrument(instrument_name, *args, **kwargs)
    except Exception as exc:
        device_health_hub.register_failure(
            instrument_key=instrument_name,
            model=instrument_name,
            device_type="Unknown",
            serial_number=serial_hint or "Unknown",
            error_message=str(exc),
            unrecoverable=True,
        )
        raise

    model, device_type, serial_number = _extract_device_info(instrument_name, device)
    device_health_hub.register_connection(
        instrument_key=instrument_name,
        model=model,
        device_type=device_type,
        serial_number=serial_number,
    )
    _attach_runtime_hooks(
        instrument_name=instrument_name,
        device=device,
        model=model,
        device_type=device_type,
        serial_number=serial_number,
    )
    return device


def PM100D(device_sn: str = "") -> PM100DDriver:
    """
    Create a Thorlabs PM100D power meter controller.

    Parameters
    ----------
    device_sn : str, optional
        Optional serial selector, by default ``""``.

    Returns
    -------
    PM100DDriver
        PM100D instrument object.
    """
    return cast("PM100DDriver", _build_instrument("PM100D", device_sn))


def BSC203_HDR50() -> BSC203_HDR50Driver:
    """
    Create a Thorlabs BSC203 + HDR50 rotation stage controller.

    Returns
    -------
    BSC203_HDR50Driver
        BSC203_HDR50 instrument object.
    """
    return cast("BSC203_HDR50Driver", _build_instrument("BSC203_HDR50"))


def PM400(device_sn: str = "") -> PM400Driver:
    """
    Create a Thorlabs PM400 power meter controller.

    Parameters
    ----------
    device_sn : str, optional
        Optional serial selector, by default ``""``.

    Returns
    -------
    PM400Driver
        PM400 instrument object.
    """
    return cast("PM400Driver", _build_instrument("PM400", device_sn))


def DC2200(device_sn: str = "") -> DC2200Driver:
    """
    Create a Thorlabs DC2200 LED controller.

    Parameters
    ----------
    device_sn : str, optional
        Optional serial selector, by default ``""``.

    Returns
    -------
    DC2200Driver
        DC2200 instrument object.
    """
    return cast("DC2200Driver", _build_instrument("DC2200", device_sn))


def M69920(device_sn: str = "") -> M69920Driver:
    """
    Create a Newport M69920 arc lamp power supply controller.

    Parameters
    ----------
    device_sn : str, optional
        Optional serial selector, by default ``""``.

    Returns
    -------
    M69920Driver
        M69920 instrument object.
    """
    return cast("M69920Driver", _build_instrument("M69920", device_sn))


def CS260B(device_sn: str = "") -> CS260BDriver:
    """
    Create a Newport CS260B monochromator controller.

    Parameters
    ----------
    device_sn : str, optional
        Optional serial selector, by default ``""``.

    Returns
    -------
    CS260BDriver
        CS260B instrument object.
    """
    return cast("CS260BDriver", _build_instrument("CS260B", device_sn))


def HR4000CG(device_sn: str = "") -> HR4000CGDriver:
    """
    Create an OceanOptics HR4000CG spectrometer controller.

    Parameters
    ----------
    device_sn : str, optional
        Optional serial selector, by default ``""``.

    Returns
    -------
    HR4000CGDriver
        HR4000CG instrument object.
    """
    return cast("HR4000CGDriver", _build_instrument("HR4000CG", device_sn))


def RS_7_1(device_sn: str = "", power_on_test: bool = True) -> RS_7_1Driver:
    """
    Create a Gamma Scientific RS-7-1 tunable light source controller.

    Parameters
    ----------
    device_sn : str, optional
        Optional serial selector, by default ``""``.
    power_on_test : bool, optional
        Whether to run power-on integrity tests, by default True.

    Returns
    -------
    RS_7_1Driver
        RS_7_1 instrument object.
    """
    return cast("RS_7_1Driver", _build_instrument("RS_7_1", device_sn, power_on_test))


def SP_2150(device_sn: str = "") -> SP_2150Driver:
    """
    Create a Princeton Instruments SP-2150 monochromator controller.

    Parameters
    ----------
    device_sn : str, optional
        Optional serial selector, by default ``""``.

    Returns
    -------
    SP_2150Driver
        SP_2150 instrument object.
    """
    return cast("SP_2150Driver", _build_instrument("SP_2150", device_sn))


def USB_520(
    device_sn: str = "",
    tear_on_startup: bool = True,
    reverse_negative: bool = True,
) -> USB520Driver:
    """
    Create a Futek USB-520 load-cell ADC controller.

    Parameters
    ----------
    device_sn : str, optional
        Device serial selector or channel alias, by default ``""``.
    tear_on_startup : bool, optional
        Run tare calibration at startup, by default True.
    reverse_negative : bool, optional
        Flip sign for negative-force orientation, by default True.

    Returns
    -------
    USB520Driver
        USB_520 instrument object.
    """
    return cast(
        "USB520Driver",
        _build_instrument(
            "USB_520",
            device_sn,
            tear_on_startup=tear_on_startup,
            reverse_negative=reverse_negative,
        ),
    )


def mantisCam(device_sn: str = "") -> MantisCamCtrl:
    """
    Create a MantisCam ZMQ control client.

    Parameters
    ----------
    device_sn : str, optional
        Optional logical camera identifier, by default ``""``.

    Returns
    -------
    MantisCamCtrl
        MantisCam control instrument object.

    Notes
    -----
    Camera family detection (GSense/F13/etc.) is derived from live metadata
    published by MantisCamUnified; no manual type flag is required.
    """
    return cast(
        "MantisCamCtrl",
        _build_instrument(
            "mantisCam",
            device_sn,
            log_level=__GLOBAL_LOG_LEVEL,
        ),
    )


__all__ = [
    "init_logger",
    "PM100D",
    "BSC203_HDR50",
    "PM400",
    "DC2200",
    "M69920",
    "CS260B",
    "HR4000CG",
    "RS_7_1",
    "SP_2150",
    "USB_520",
    "mantisCam",
]



from ..interfaces._bsl_serial import _bsl_serial as bsl_serial
from ..headers._bsl_inst_info import _bsl_inst_info_list as inst
from ..headers._bsl_logger import _bsl_logger as bsl_logger
from ..headers._bsl_type import _bsl_type as bsl_type

import time
import enum
import numpy as np
from typing import Union
from numpy.typing import NDArray

from loguru import logger

try:
    from pycolorname.pantone.pantonepaint import PantonePaint
except Exception:
    PantonePaint = None

try:
    from skimage import color
except Exception:
    color = None

class RS_7_1:
    CONNECT_RETRY_COUNT = 3
    CONNECT_RETRY_DELAY_SEC = 1.0

    class _SYSTEM_UNIT(enum.Enum):
        RADIOMETRIC = 0; PHOTOMETRIC = 1; PERCENTAGE = 2
    class _STM_MODE(enum.Enum):
        ASCII_COMMA = 0; ASCII_COLUMN = 1; PACKED_BINARY = 2
    class POWER_UNIT(enum.Enum):
        UNITS=['uW/cm^2/sr','uW/cm^2','nits','lux','%']
        RADIANCE = 0; IRRADIANCE = 1; LUMINANCE = 2; ILLUMINANCE=3; PERCENTAGE=4
    class OBSERVER_ANGLE(enum.Enum):
        DEG_2 = 2; DEG_10 = 10
    class LED_CHANNELS(enum.Enum):
        LEN_CHANS = list([3,4,5,6,7,8,11,13,14,16,19,20,21,22,23,24,26,27,28,29,30,31,33,34,36,37,38,39,41,42,43,45,46,47,49,51,52,53,54,55,57,59,60,61,62,63])
        LEN_CHANS_NO_WHITE = list([3,4,5,6,7,8,11,13,14,16,19,20,21,22,23,24,26,27,28,29,33,34,36,39,41,42,43,45,46,47,49,51,52,53,54,55,57,59,60,61])
        FWHM = list([0.0, 0.0, 13.62, 27.44, 13.62, 16.41, 30.97, 14.81, 0.0, 0.0, 22.25, 0.0, 18.07, 24.06, 0.0, 14.81, 0.0, 0.0, 35.15, 106.8, 106.8, 32.22, 25.21, 18.53, 0.0, 20.28, 79.39, 79.39, 24.06, 13.62, 0.0, 0.0, 40.32, 17.83, 0.0, 21.05, 16.15, 0.0, 33.17, 0.0, 19.7, 24.06, 21.86, 0.0, 24.06, 31.81, 16.94, 0.0, 21.27, 0.0, 20.94, 29.85, 52.46, 21.36, 21.36, 0.0, 18.53, 0.0, 15.1, 21.86, 27.68, 31.81, 0.0, 0.0])
        WAVELENGTH = list([0.0, 0.0, 590.35, 498.75, 590.35, 399.0, 521.85, 627.11, 0.0, 0.0, 769.71, 0.0, 657.0, 712.89, 0.0, 627.11, 0.0, 0.0, 845.9, 571.15, 571.15, 901.51, 746.37, 632.75, 0.0, 452.86, 610.19, 610.19, 712.89, 590.35, 5990.9, 0.0, 936.91, 426.01, 0.0, 688.43, 616.27, 2937.8, 531.37, 0.0, 445.77, 729.16, 495.49, 0.0, 729.16, 525.64, 667.09, 0.0, 407.85, 0.0, 753.59, 474.73, 959.3, 700.74, 700.74, 0.0, 632.75, 0.0, 426.8, 495.49, 802.68, 525.64, 2747.6, 0.0])
    class MACBECH_COLOR(enum.Enum):
        CIExy = list([(0.4336, 0.3787), (0.4187, 0.3749), (0.2757, 0.2996), (0.3688, 0.4501), (0.3016, 0.2871), (0.2856, 0.3905), (0.5291, 0.4081), (0.2335, 0.2157), (0.5002, 0.3295), (0.3316, 0.2544), (0.3986, 0.5002), (0.4960, 0.4426), (0.2042, 0.1676), (0.3262, 0.5040), (0.5734, 0.3284), (0.4693, 0.4730), (0.4175, 0.2702), (0.2146, 0.3028), (0.3486, 0.3625)])
    def __init__(self, device_sn="", power_on_test:bool = True) -> None:
        """
        Initialize and connect an RS-7-1 tunable light source.

        Parameters
        ----------
        device_sn : str, optional
            Optional serial selector, by default ``""``.
        power_on_test : bool, optional
            Run startup integrity tests, by default True.
        """
        self._target_device_sn = device_sn
        self.inst = inst.RS_7_1
        self.device_id = ""
        self.logger = bsl_logger(self.inst)
        self.logger.info(f"Initiating bsl_instrument - RS_7_1({device_sn})...")
        self._pantone_keys = list(PantonePaint().keys()) if PantonePaint is not None else []
        self._com = self._serial_connect()

        if self._com is not None:
            if self._com.serial_port is not None:
                self.device_id = self._com_query('USN')
                self.logger.device_id = self.device_id
                self._system_init(power_on_test = power_on_test)
                self.logger.success(f"READY - Tunable Light Source.\n\n\n")
                return None
        self.logger.error(f"FAILED to connect to RS_7_1 Tunable Light Source!\n\n\n")
        raise bsl_type.DeviceConnectionFailed

    def __del__(self, *args, **kwargs) -> None:
        try:
            self.close()
        except Exception:
            pass
        return None

    def _serial_connect(self) -> bsl_serial:
        """
        Connect to RS-7-1 serial interface.

        Returns
        -------
        bsl_serial | None
            Serial interface instance when successful.
        """
        com_port = None
        try:
            com_port = bsl_serial(inst.RS_7_1, self._target_device_sn)
        except Exception as e:
            self.logger.error(f"{type(e)}")
        return com_port

    def reconnect(self, device_sn: str = "", power_on_test: bool = False) -> bool:
        """
        Reconnect to RS-7-1 and optionally re-run startup tests.

        Parameters
        ----------
        device_sn : str, optional
            Optional serial selector override, by default ``""``.
        power_on_test : bool, optional
            Run integrity/basic assurance tests after reconnect, by default False.

        Returns
        -------
        bool
            True when reconnection succeeds.
        """
        if device_sn:
            self._target_device_sn = device_sn
        self.close()
        last_error = None
        for attempt in range(1, self.CONNECT_RETRY_COUNT + 1):
            try:
                self._com = self._serial_connect()
                if self._com is None or getattr(self._com, "serial_port", None) is None:
                    raise bsl_type.DeviceConnectionFailed("No serial communication port found.")
                self.device_id = self._com_query("USN")
                self.logger.device_id = self.device_id
                self._system_init(power_on_test=power_on_test)
                return True
            except Exception as exc:
                last_error = exc
                self.logger.warning(
                    f"Reconnect attempt {attempt}/{self.CONNECT_RETRY_COUNT} failed: {type(exc)}"
                )
                time.sleep(self.CONNECT_RETRY_DELAY_SEC)
        self.logger.error(f"Unable to reconnect RS_7_1: {repr(last_error)}")
        return False

    def reset_system(self, power_on_test: bool = False) -> bool:
        """
        Reboot RS-7-1 and reinitialize runtime settings.

        Parameters
        ----------
        power_on_test : bool, optional
            Run startup tests after reboot, by default False.

        Returns
        -------
        bool
            True when reset succeeds.
        """
        try:
            self._system_restart()
            self._system_init(power_on_test=power_on_test)
            return True
        except Exception as exc:
            self.logger.error(f"Failed to reset RS_7_1: {type(exc)}")
            return False

    #checked
    def _system_restart(self) -> None:
        """
        Reboot the light source to reset all runtime variables.
        """
        self.logger.info("Rebooting System!")
        self._com.writeline('RST')
        time.sleep(5)
        self.logger.info("System Rebooted!")
        return None

    def _system_init(self, power_on_test: bool=True) -> None:
        if power_on_test:
            try:
                self._system_restart()
            except Exception as exc:
                self.logger.warning(
                    "RS_7_1 restart during power-on test failed (continuing): {}",
                    exc,
                )

            try:
                self._run_integrity_check()
            except Exception as exc:
                self.logger.warning(
                    "RS_7_1 integrity check failed (warning-only, continuing): {}",
                    exc,
                )

            try:
                self._run_basic_assurance_test()
            except Exception as exc:
                self.logger.warning(
                    "RS_7_1 basic assurance test failed (warning-only, continuing): {}",
                    exc,
                )
        self._set_wavelength_range(360,1100)
        self.set_standard_observer_angle(self.OBSERVER_ANGLE.DEG_2)
        self._set_spectrum_transfer_format(self._STM_MODE.ASCII_COMMA)
        self.set_iris_position(0)
        return None

    def _run_integrity_check(self) -> None:
        """
        - Performs an internal check of all RS-7 parameters and calibration 
        data, and verifies that:
            a) all monitored internal power supply rail voltages 
                are within tolerance.
            b) all parameters are within range & structures are intact.
            c) stored calibration data matches the serial number of the 
                LED board for which they were generated d) all structures 
                have a valid CRC (CRC32).

        Raises:
        --------
            Failed Test : `bsl_type.DeviceInconsistentError`
                System failed its power-on integrity test.
        """
        if self._com_cmd('ICK',5) != 0:
            self.logger.error("Failed power-on integrity check!")
            raise bsl_type.DeviceOperationError
        else:
            self.logger.success(f'Passed power-on integrity check.')
        return None

    def _run_basic_assurance_test(self) -> None:
        """
        - Rapidly sequence through all channels in order to verify basic functionality. 
        This is intended to execute as quickly as possible and is not intended as a 
        comprehensive audit of each channel’s performance. 

        - The test criteria is that each channel, set at 50% of maximum power, 
        demonstrates output power of at least 90% of the expected value.
        
        Raises:
        --------
            Faild Test : `bsl_type.DeviceInconsistentError`
                System failed its power-on integrity test.
        """
        self._com_cmd('BAT',5)
        if self._com_cmd('ICK',5) != 0:
            self.logger.error("failed power-on basic assurance test!")
            raise bsl_type.DeviceOperationError
        else:
            self.logger.success(f'    Passed power-on basic assurance test.')
        return None

    def _set_UNI_unit(self, unit:_SYSTEM_UNIT) -> None:
        """
        - Choose the system UNI power unit.

        Parameters
        -----------
        unit : `RS_7_1._SYSTEM_UNIT`
            Choose from _SYSTEM_UNIT.RADIOMETRIC, _SYSTEM_UNIT.PHOTOMETRIC, 
            and _SYSTEM_UNIT.PERCENTAGE
        """
        msg = f"UNI{unit}"
        self._com_cmd(msg)
        return None
    
    def _set_irr_distance(self, distance_mm:int=0) -> None:
        """
        - Set the imaging/output plane's distance from the output port
        of the light source. (i.e. front surface)

        - In conjunction with _set_UNI_unit(), one can dictates the power
        operation unit of the system from [radiance, irradiance, luminance, 
        illuminance, and percentage].

        - When set to 0, indicating system running in luminance/radiance mode.

        Parameters
        -----------
        distance_mm : `int`
            (default = 0)
            distance from the output port of the light source in mm.
        """
        msg = f"IRR{str(distance_mm)}"
        self._com_cmd(msg)
        return None

    def _set_power_unit(
        self, unit:POWER_UNIT, irr_distance_mm:int = 0) -> None:
        
        if unit==self.POWER_UNIT.PERCENTAGE:
            self._set_UNI_unit(self._SYSTEM_UNIT.PERCENTAGE.value)
            self._set_irr_distance(0)
        elif unit==self.POWER_UNIT.RADIANCE:
            self._set_UNI_unit(self._SYSTEM_UNIT.RADIOMETRIC.value)
            self._set_irr_distance(0)
        elif unit==self.POWER_UNIT.IRRADIANCE:
            self._set_UNI_unit(self._SYSTEM_UNIT.RADIOMETRIC.value)
            self._set_irr_distance(irr_distance_mm)
        elif unit==self.POWER_UNIT.LUMINANCE:
            self._set_UNI_unit(self._SYSTEM_UNIT.PHOTOMETRIC.value)
            self._set_irr_distance(0)
        elif unit==self.POWER_UNIT.ILLUMINANCE:
            self._set_UNI_unit(self._SYSTEM_UNIT.PHOTOMETRIC.value)
            self._set_irr_distance(irr_distance_mm)

        self.logger.debug(f"Power Unit set to {unit} with imaging distance = {irr_distance_mm}mm.")
        return None

    def _set_optical_feedback(self, FBK:bool=True) -> None:
        """
        - Enable/Disable the realtime optical feedback that ultilize
        an internal photodiode in the light source to dynamically matching
        output power to requested power instead of relying on computed power.

        Parameters
        -----------
        FBK : `bool`
            (default = Ture)
        """
        msg = f"FBK{int(FBK)}"
        self._com_cmd(msg)
        self.logger.info(f"Optical Feedbacl set to {FBK}")
        return None

    def _set_spectrum_transfer_format(self, mode:_STM_MODE = _STM_MODE.ASCII_COMMA) -> None:
        """
        - WARNING: Changing wavelength_range away from default will
        break the normal function of other provided functions such as
        `get_current_spectrum`, `set_spectrum_raw` etc.,

        Parameters
        -----------
        mode : `_STM_MODE`
            (default = _STM_MODE.ASCII_COMMA)
        """
        self._com_cmd(f"STM{mode.value}")
        self.logger.info(f"Spectrum transfer format set to {mode}.")
        return None

    def _set_wavelength_range(self, min:int=360, max:int=1100) -> None:
        """
        - Set the wavelength range for OSP (Output Specturm) and 
        TSP (Set Target Spectrum) operations.

        - WARNING: Changing wavelength_range away from default will
        break the normal function of other provided functions such as
        `get_current_spectrum`, `set_spectrum_raw` etc.,


        Parameters
        ----------
        min : `int`
            (default to 360)
            Minimum wavelength for spectrum operation in nm.
            
        max : `int`
            (default to 1100)
            Maximum wavelength for spectrum operation in nm.
        """
        self._wavelength_min = min
        self._wavelength_max = max
        self._com_cmd(f"WLR{min},{max}")
        self.logger.info(f"Spectrum operation wavelength range set to {min}nm to {max}nm.")
        return None

    def _black_body_spectrum(self, temp:int) -> 'list[float]':
        spec = [self._planck(x+360,temp) for x in range(741)]
        return spec

    def _planck(self, wav:Union[float, int], T:int) -> float:
        h = 6.626e-34
        c = 3.0e+8
        k = 1.38e-23
        wav = wav*1e-9
        a = 2.0*h*c**2
        b = h*c/(wav*k*T)
        intensity = a/ ( (wav**5) * (np.exp(b) - 1.0) )
        return intensity

    def _com_query(self, msg, timeout:float = 0.5) -> str:
        self._com.flush_read_buffer()
        self._com.write(msg+'\r\n')
        resp = self._com.readline()
        if resp == "":
            return self._com.readline()
        else:
            return resp

    def _com_cmd(self, msg, timeout:float = 0.5) -> int:
        self._com.flush_read_buffer()
        self._com.write(msg+'\r\n')
        resp = self._com.readline()
        if resp == "":
            resp = self._com.readline()
            if resp != "Ok":
                self.logger.error(f"message from device: \"{resp}\"")
                raise bsl_type.DeviceOperationError
            else:
                return 0
        elif resp != "Ok":
            self.logger.error(f"message from device: \"{resp}\"")
            raise bsl_type.DeviceOperationError

    #checked
    def find_closest_chan(self, wavelength:float) -> 'tuple([list[int], float, float])':
        """
        Return the channels' number, its[their] corresponding wavelength,
        and the FWHM of the LED channel[s] with closest wavelength provided.

        Uses
        ---------
        >>> led_chans = light.find_closest_chan(wavelength=500)[0]
        >>> light.set_power_chans(led_chans, powers=40)

        Parameters
        -----------
        wavelength : `float`
            Wavelength of interest to be accquired.

        Returns
        ----------
        LED_index[es] : `list[int]`
            List of index/indexes of LED channels that match closest to the
            requrested wavelength in the parameter.
        
        LED_wavelength : `float`
            Wavelength of the LED[s] with the channel specified in the LED_index[es].

        LED_FWHM : `float`
            FWHM of the the LED[s] with the channel specified in the LED_index[es].
        """
        lst = np.asarray(self.LED_CHANNELS.WAVELENGTH.value)
        fwhm = self.LED_CHANNELS.FWHM.value
        idx = (np.abs(lst - wavelength)).argmin()
        list_idx = list(np.where(lst==lst[idx])[0]+1)
        self.logger.debug(f"Found LED ({repr(list_idx)}) with wavelength {lst[idx]}nm, FWHM {fwhm[idx]}nm from requested wavelength {wavelength}nm.")
        return (list_idx, lst[idx], fwhm[idx])

    #checked
    def set_iris_position(self, percentage:int=0) -> None:
        """
        - Set iris position as percentage **closed** (0 ~ 100%).
        - Make sure to set iris before seting output power in irradiance mode.

        Parameters:
        -----------
        percentage : `int`
            (Defaults to 0)
            Percentage close, i.e. to fully open the iris, set to 0.
        """
        if percentage < 0 or percentage > 100:
            self.logger.error("Cannot set percentage smaller than 0 or greater than 100!")
            raise bsl_type.DeviceOperationError
        msg = f"IRI{percentage}"
        self._com_cmd(msg)
        self.logger.info(f"Iris position set to {percentage}% closed.")
        time.sleep(3)
        return None

    #checked
    def set_standard_observer_angle(
        self, angle:OBSERVER_ANGLE=OBSERVER_ANGLE.DEG_2) -> None:
        """
        - Sets the CIE Tristimulus Standard Observer Angle as 2 degrees or 10 degrees. 
        The default at power-on is always 2 degrees.

        Parameters
        ----------
        angle : `RS_7_1.OBSERVER_ANGLE`
            (default to DEG_2)
            CIE Tristimulus Standard Observer Angle.
        """
        msg = f"SOB{angle.value}"
        self._com_cmd(msg)
        self.logger.debug(f"Standard Observer Angle set to {angle.value} degrees")
        return None


    #checked
    def set_power_all(
        self, power:float, unit:POWER_UNIT=POWER_UNIT.PERCENTAGE, 
        irr_distance_mm:int=0) -> None:
        """
        - Set the output power of the overall system to specified unit and power.

        Parameters:
        ----------
        power : `float`
            Power of all channel in percentage.

        unit : `RS_7_1.POWER_UNIT`
            (default to Percentage)
            Select the unit of the power from radiance, irradiance, luminance,
            illuminance, and percentage.

        irr_distance_mm : `int`
            (default to 0)
            Distance from the surface of the output port of the light source to
            the desired imaging plane. Only used for irradiance or illuminance
            power profile.
        """
        self._set_power_unit(unit, irr_distance_mm)
        msg = f"SCP 0,{power}"
        self._com_cmd(msg)
        self.logger.info(f"All LED channels' power set to {power}{self.POWER_UNIT.UNITS.value[unit.value]}.")
        return None

    #checked
    def set_power_chans(self, chans:Union['list[int]', NDArray[np.int_], int], powers:Union['list[float]', NDArray[np.float64], float], unit:POWER_UNIT=POWER_UNIT.PERCENTAGE, irr_distance_mm:int=0) -> None:
        """
        - Set the output power of the individual LED channel[s] tothe 
        specified unit and power.

        Parameters
        ----------
        chans : `list[int]` or 'int' or 'NDArray[int]'
            A int list consists individual LED channel number with an
            one-to-one correspondence to the powers array.

        powers : `list[float]` or 'float' or 'NDArray[float]'
            A floating number NDArray consists individual power settings with 
            an one-to-one correspondence to the chans array.

        unit : `RS_7_1.POWER_UNIT`
            (default to Percentage)
            Select the unit of the power from radiance, irradiance, luminance,
            illuminance, and percentage.

        irr_distance_mm : `int`
            (default to 0)
            Distance from the surface of the output port of the light source to
            the desired imaging plane. ONLY used for irradiance or illuminance
            power profile.
        """
        self._set_power_unit(unit, irr_distance_mm)
        if type(powers) is np.ndarray:
            powers = list(powers) 
        if type(chans) is np.ndarray:
            chans = list(chans)
        if type(powers) is not list:
            powers = list([powers]) 
        if type(chans) is not list:
            chans = list([chans]) 
        if len(powers)==1:
            powers = powers * len(chans)

        if (len(chans)!=len(powers)):
            self.logger.error("Provided list of channels doesn't have same amount of elemets as the list of power!")
            raise bsl_type.DeviceOperationError

        for chan in chans:
            if (chan not in self.LED_CHANNELS.LEN_CHANS.value):
                self.logger.error(f"Provided LED channel {chan} is not installed!")
                raise bsl_type.DeviceOperationError
        
        str_chans = [str(x) for x in chans]
        str_powers = [str(f'{x:.4f}') for x in powers]
        combined = str_chans+str_powers
        combined[0::2] = str_chans
        combined[1::2] = str_powers
        combined_proc = ','.join(combined)
        msg = f"SCP{combined_proc}"
        self.logger.info("LED Channel[s] power set.")
        self._com_cmd(msg)

    #checked
    def set_power_led_random(self, power_percentage:int=5) -> 'list[float]':
        """
        - Set output spectrum to a random spectrum.

        Uses
        ----------
        >>> set_spectrum_random()
        >>> set_power_output(power = 30)

        Returns
        --------
        Output_spectrum : `list[float]`
            Actual output spectrum with a.u. from 360nm to 1100nm with step size of 1nm.
        """
        spectrum = np.random.random([len(self.LED_CHANNELS.LEN_CHANS_NO_WHITE.value)])*power_percentage
        self.set_power_chans(self.LED_CHANNELS.LEN_CHANS_NO_WHITE.value,spectrum)
        self.logger.info("Output spectrum set to a randomly generated spectrum.")
        return None

    #checked
    def set_power_output(
        self, power:float, unit:POWER_UNIT=POWER_UNIT.RADIANCE, 
        irr_distance_mm:int=0, match_chrom:bool=False) -> None:
        """
        - Set the output power of the overall system to specified unit and power.

        - WARNINGL: Unknown chromaticity/spectrum accuracy! Always use in 
        conjunction with `get_power_output()` and `get_spectrum_output()` functions 
        or with external powermeter and spectrometer! For better chromaticity/
        spectrum accuracy, set your desired absolute (photometric/radiometric) unit 
        in the `set_spectrum_raw()` functions! 
        
        - WARNING: For chromaticity/color based spectrum setting, follow instructions 
        in `set_spectrum_rgb()`, `set_spectrum_CIExy()`, and `set_spectrum_pantone()` 
        to ensure chromaticity precision!

        - WARNING: Using `match_chrom=True` may results in faliure to change the power 
        level since the chromaticity correction function may lower the output level 
        for unknown reason, NO warning NOR error will be raised if this happened!

        Parameters
        ----------
        power : `float`
            Power of all channel in percentage.

        unit : `RS_7_1.POWER_UNIT`
            (default to Radiance)
            Select the unit of the power from radiance, irradiance, luminance,
            illuminance, and percentage.

        irr_distance_mm : `int`
            (default to 0)
            Distance from the surface of the output port of the light source to
            the desired imaging plane. Only used for irradiance or illuminance
            power profile.

        match_chrom : `bool`
            (default to False)
            Note that changing the output level using `match_chrom = False` can 
            result in a chromaticity shift in the final output spectrum OSP, as 
            OUT merely adjusts all channels’ power level to match the requested 
            setting without regard to any wavelength shift caused by the change 
            in drive current.
        """
        self._set_power_unit(unit, irr_distance_mm)
        chrom = self.get_chromaticity_output()
        if match_chrom:
            chrom = self.get_chromaticity_output()
            msg = f'OUT{power:.3f}'
            self._com_cmd(msg)
            self._com_cmd(f"CCS{chrom[0]:.6f},{chrom[1]:.6f}")
        else:
            msg = f'OUT{power:.3f}'
            self._com_cmd(msg)
        self.logger.info(f"Output power set to: {power}{self.POWER_UNIT.UNITS.value[unit.value]}, irradiance distance: {irr_distance_mm}, chroma_correction: {match_chrom}.")
        return None

    #checked
    def set_power_fixed_spectrum(
        self, power:float, unit:POWER_UNIT=POWER_UNIT.RADIANCE, 
        irr_distance_mm:int=0) -> None:
        """
        - Set the output power of the overall system to specified unit and power
        while maintaining the current output spectrum.

        Parameters
        ----------
        power : `float`
            Power of all channel in percentage.

        unit : `RS_7_1.POWER_UNIT`
            (default to Radiance)
            Select the unit of the power from radiance, irradiance, luminance,
            illuminance, and percentage.

        irr_distance_mm : `int`
            (default to 0)
            Distance from the surface of the output port of the light source to
            the desired imaging plane. Only used for irradiance or illuminance
            power profile.
        """
        self.set_spectrum_raw(self.get_spectrum_output(),power=power,power_unit=unit,irr_distance_mm=irr_distance_mm)
        self.logger.info(f"Output spectrum power set to: {power}{self.POWER_UNIT.UNITS.value[unit.value]}, irradiance distance: {irr_distance_mm}.")
        return None

    #checked
    def set_spectrum_raw(
        self, 
        spectrum:Union['list[float]', NDArray[np.float64]], 
        *,
        power:float=0, 
        power_unit:POWER_UNIT=POWER_UNIT.RADIANCE, 
        include_white:bool=True, 
        fit_max_pwr:bool=False, 
        chroma_correction:bool = False,
        irr_distance_mm:int=0) -> float:
        """
        - Sets and fit the spectrum to be fitted by the light source, with 1nm step size, and unit
        of radiance or irradiance ONLY. Default range from 360nm to 1100nm i.e. 741 points.

        - For lot of applications, it's easier to provide a normalized spectrum with a.u. 
        and set the arg. "fit_max_pwr" to Ture to get maximum power output for a specific
        spectrum of interest.

        Parameters
        ----------
        spectrum : `list[float]` or `NDArray[float]`
            Specturm data in specified unit with 1nm step size.

        power : `float`
            (default = 100)
            A floating number of individual power settings.
        
        power_unit : 'RS_7_1.POWER_UNIT'
            (defalut = POWER_UNIT.RADIANCE)
        
        include_white : `bool`
            (defalut = True)
            Whether including white LEDs in the fitting process.
        
        fit_max_pwr : `bool`
            (defalut = False)
            Fit to maximum possible power of the light source or not.
        
        chroma_correction : `bool`
            (defalut = False)
            Whether run chromaticity correction to the fitted spectrum to the 
            chromaticity of the request spectrum.

        irr_distance_mm : `int`
            (default to 0)
            Distance from the surface of the output port of the light source to
            the desired imaging plane. Only used for irradiance or illuminance
            power profile.

        Returns
        --------
        RMS_Error : `float`
            Root-Mean-Square error for the fitted specturm vs. provided spectrum.
        """
        if (power_unit is self.POWER_UNIT.PERCENTAGE):
            self.logger.error("Only Radiometric and Photometric are supported for spectrum setting!")
            raise bsl_type.DeviceOperationError
        self._set_power_unit(power_unit, irr_distance_mm)
        
        spectrum = list(spectrum)
        if len(spectrum) != (self._wavelength_max - self._wavelength_min + 1):
            self.logger.error("Provided spectrum data's length doesn't match current wavelength min_max setting!")
            raise bsl_type.DeviceOperationError
        msg_spectrum = ','.join(['{:.6f}'.format(x) for x in spectrum])
        self._com_cmd(f"TSP{msg_spectrum}")
        
        if power != 0:
            self._com_cmd(f"STS{power:.4f}")
            self.logger.debug(f"Target spectrum power scaled to {power}{self.POWER_UNIT.UNITS.value[power_unit.value]}.")

        msg_cmd = "FTS"
        if include_white:
            msg_cmd = msg_cmd + "W"
        if fit_max_pwr:
            msg_cmd = msg_cmd + "M"

        self._com_cmd(f"{msg_cmd}")
        if chroma_correction:
            self._com_cmd("CCS")
        return self.get_E_rms_fitted_spectrum()

    #checked
    def set_spectrum_CIExy(
        self, CIEx:float, CIEy:float,
        power:float=0, power_unit:POWER_UNIT=POWER_UNIT.RADIANCE, irr_distance_mm:int=0
        ) -> 'tuple[float, float]':
        """
        - Fit the output spectrum to a specified CIE 1931 x,y chromaticity setting.
        
        - Set `match_chrom = Ture` when calling set_`power_fitted_spectrum()`
        to maintain the chromaticity matching. 

        Uses
        ----------
        >>> set_CIE_chroma(0.125, 0.246)
        >>> set_power_output(20, match_chrom=True)

        Parameters
        ----------
        CIEx : 'float'
            CIEx coordinate for chromaticity setting.

        CIEy : 'float'
            CIEy coordinate for chromaticity setting.

        Returns:
        ----------
        (A_CIEx, A_CIEy) : `[float, float]`
            Actual fitted CIEx,y chromaticity in CIE 1931 standard.
        """
        self.set_power_all(0.1)
        self._com_cmd(f"CCS{CIEx:.6f},{CIEy:.6f}")
        if power !=0:
            self.set_power_output(power, power_unit, irr_distance_mm)
        self._com_cmd(f"CCS{CIEx:.6f},{CIEy:.6f}")
        self.logger.debug(f"Set output spectrum to CIExy chromaticity {CIEx:.6f},{CIEy:.6f}.")
        return self.get_chromaticity_output()
        

    #checked
    def set_spectrum_black_body(
        self, temp:int, power:float=0, power_unit:POWER_UNIT=POWER_UNIT.RADIANCE, 
        irr_distance_mm:int=0) -> float:
        """
        - Sets and fit the spectrum to be fitted by the light source, with 
        1nm step size, and unit of radiance or irradiance ONLY. Default 
        range from 360nm to 1100nm i.e. 741 points.
        
        Uses
        ----------
        >>> set_specturm_black_body(temp=3000,30)

        Parameters
        ----------
        temp : `int`
            Requested black body temperature in Kelvins.
        
        power_unit : `RS_7_1.POWER_UNIT`
            (default to radiance)
            Select the unit of the power from radiance, irradiance, luminance,
            illuminance. NO percentage setting available for spectrum based 
            output setting! 

        irr_distance_mm : `int`
            (default to 0)
            Distance from the surface of the output port of the light source to
            the desired imaging plane. ONLY used for irradiance or illuminance
            power profile.

        Returns
        --------
        Color Temperature : `float`
            Output color temperature in Kelvins.
        """
        spectrum = self._black_body_spectrum(temp)
        xmin=min(spectrum) 
        xmax=max(spectrum)
        for i, x in enumerate(spectrum):
            spectrum[i] = (x-xmin) / (xmax-xmin)
        
        self.logger.info(f"Setting Output spectrum to Black Body spectrum with temperature: {temp} Kelvins.")
        self.set_spectrum_raw(spectrum, power=power, power_unit=power_unit, irr_distance_mm=irr_distance_mm)
        return self.get_color_temp()

    def set_spectrum_hsv(
        self, H:float, S:float, V:float,
        power:float=0, power_unit:POWER_UNIT=POWER_UNIT.RADIANCE, irr_distance_mm:int=0
        ) -> float:
        """
        - Set output spectrum to a specified HSV color profile.
        
        - Set `match_chrom = Ture` when calling set_`power_fitted_spectrum()`
        to maintain the chromaticity matching. 

        Uses
        ----------
        >>> set_spectrum_hsv(125,250,85,400)
        >>> set_power_output(20, match_chrom=True)

        Parameters
        ----------
        H : `float`
            Hue from 0 to 360 degrees.

        S : `float`
            Saturation from 0 to 100 percent.

        V : `float`
            Value from 0 to 100.
        
        Returns
        --------
        RMS_Error : `float`
            Root-Mean-Square error for the fitted specturm.
        """

        # Convert saturation and value to a range of 0 to 1
        S /= 100.0
        V /= 100.0
        # Calculate the chroma
        chroma = V * S

        # Find the position within the hexagon
        hue_prime = H / 60.0
        x = chroma * (1 - abs(hue_prime % 2 - 1))

        # Compute the intermediate RGB values based on the hue_prime
        if hue_prime < 1:
            r1, g1, b1 = chroma, x, 0
        elif hue_prime < 2:
            r1, g1, b1 = x, chroma, 0
        elif hue_prime < 3:
            r1, g1, b1 = 0, chroma, x
        elif hue_prime < 4:
            r1, g1, b1 = 0, x, chroma
        elif hue_prime < 5:
            r1, g1, b1 = x, 0, chroma
        else:
            r1, g1, b1 = chroma, 0, x

        # Match the luminance by adding the same amount to each component
        m = V - chroma
        r, g, b = r1 + m, g1 + m, b1 + m

        # Convert to 8-bit values
        r, g, b = int(r * 255), int(g * 255), int(b * 255)

        return self.set_spectrum_rgb(r,g,b,power,power_unit,irr_distance_mm)
    

    #checked
    def set_spectrum_rgb(
        self, r:np.uint8, g:np.uint8, b:np.uint8,
        power:float=0, power_unit:POWER_UNIT=POWER_UNIT.RADIANCE, irr_distance_mm:int=0
    ) -> float:
        """
        - Set output spectrum to a specified RGB color profile.
        
        - Set `match_chrom = Ture` when calling set_`power_fitted_spectrum()`
        to maintain the chromaticity matching. 

        Uses
        ----------
        >>> set_spectrum_rgb(125,250,85,400)
        >>> set_power_output(20, match_chrom=True)

        Parameters
        ----------
        R : `int`
            RGB color Red, 8 bits color range from 0 to 255

        G : `int`
            RGB color Green, 8 bits color range from 0 to 255

        B : `int`
            RGB color Blue, 8 bits color range from 0 to 255
        
        Returns
        --------
        RMS_Error : `float`
            Root-Mean-Square error for the fitted specturm.
        """
        if ((r>255 or r<0) or (g>255 or g<0) or (b>255 or b<0)):
            self.logger.error("Provided RGB values are out of range!")
            raise bsl_type.DeviceOperationError
        if color is None:
            self.logger.error("scikit-image is required for RGB to CIExy conversion.")
            raise bsl_type.DeviceOperationError
        CIExyz = color.rgb2xyz([r/255.0,g/255.0,b/255.0])
        X=CIExyz[0]
        Y=CIExyz[1]
        Z=CIExyz[2]
        x = X / (X + Y + Z)
        y = Y / (X + Y + Z)
        self.set_spectrum_CIExy(x,y,power,power_unit,irr_distance_mm)
        self.logger.info(f"Output spectrum set to match RGB color: ({r}, {g}, {b})")
        return self.get_E_rms_fitted_spectrum()

    def set_spectrum_rgb_random(self, power:float=0, power_unit:POWER_UNIT=POWER_UNIT.RADIANCE, 
        irr_distance_mm:int=0
        ) -> float:
        """
        - Set output spectrum to a random RGB color.

        Uses
        ----------
        >>> set_spectrum_rgb_random(400)
        
        Returns
        --------
        RMS_Error : `float`
            Root-Mean-Square error for the fitted specturm.
        """
        (r,g,b) = np.random.randint(0,255,3)
        return self.set_spectrum_rgb(r,g,b,power_unit=power_unit, power=power, irr_distance_mm=irr_distance_mm)

    #checked
    def set_spectrum_pantone(
        self, color_name:str,
        power:float=0, power_unit:POWER_UNIT=POWER_UNIT.RADIANCE, irr_distance_mm:int=0
        ) -> float:
        """
        - Set output spectrum to a specified pantone color profile.

        - Set `match_chrom = Ture` when calling set_`power_fitted_spectrum()`
        to maintain the chromaticity matching. 

        Uses
        ----------
        >>> set_spectrum_pantone('Orange Peel',5)
        >>> set_power_output(20, match_chrom=True)

        Parameters
        ----------
        color_name : `str`
            Pantone Color name in `PEP8` naming convention, or a partial name
            can be provided for auto search in Pantone Color profile, a warning
            message will be raise no matter how many mathing[s] are found!
        
        Returns
        --------
        RMS_Error : `float`
            Root-Mean-Square error for the fitted specturm.
        """
        if PantonePaint is None:
            self.logger.error("pycolorname is required for Pantone color lookup.")
            raise bsl_type.DeviceOperationError
        key = [s for s in self._pantone_keys if color_name.lower() in s.lower()]
        if len(key) == 0:
            self.logger.error(f"Color \"{color_name}\" is not found in Pantone Color set!")
            raise bsl_type.DeviceOperationError
        if key[0].lower() != color_name.lower():
            self.logger.warning(f"No exact match found, assuming color \"{key[0]}\"")
        (r, g, b) = (PantonePaint()[key[0]])
        err = self.set_spectrum_rgb(r,g,b,power,power_unit,irr_distance_mm)
        self.logger.info(f"Output spectrum set to match Pantone Color: {key[0]}.")
        return err


    #checked
    def get_optical_feedback_gain(self) -> float:
        """
        Get latest optical feedback gain factor, nomnial close to 1.

        Returns:
        --------
            FBG : `float`
                Optical feedback gain
            
        """
        gain = float(self._com_query('FBG'))
        self.logger.debug(f"Current system optical feedback gain: {gain:.3f}")
        return gain

    #checked
    def get_power_output(
        self, unit:POWER_UNIT=POWER_UNIT.RADIANCE, irr_distance_mm:int = 0) -> float:
        """
        - Get the output power of the overall system in specified unit.

        Parameters
        ----------
            unit : `RS_7_1.POWER_UNIT`
                (default to Percentage)
                Select the unit of the power from radiance, irradiance, luminance,
                illuminance, and percentage.
                
            irr_distance_mm : `int`
                (default to 0)
                Distance from the surface of the output port of the light source to
                the desired imaging plane. Only used for irradiance or illuminance
                power profile.
        
        Returns
        ---------
            Actual_Power : `float`
                Actual measured output power of the light source in requested unit.
        """
        self._set_power_unit(unit, irr_distance_mm)
        cmd = f"OUTA"
        pwr=float(self._com_query(cmd))
        self.logger.debug(f"Total output power: {pwr}{self.POWER_UNIT.UNITS.value[unit.value]}.")
        return pwr

    #checked
    def get_power_all_chans(
        self, unit:POWER_UNIT=POWER_UNIT.PERCENTAGE, irr_distance_mm:int = 0
        ) -> tuple(['list[int]', 'list[float]']):
        """
        - Return all the non-OFF channels' current output powers.

        Parameters
        ----------
        unit : `RS_7_1.POWER_UNIT`
            (default to Percentage)
            Select the unit of the power from radiance, irradiance, luminance,
            illuminance, and percentage.
            
        irr_distance_mm : `int`
            (default to 0)
            Distance from the surface of the output port of the light source to
            the desired imaging plane. Only used for irradiance or illuminance
            power profile.
    
        Returns
        ---------
        (chans, powers) : `tuple(list[int], list[float])`
            Two lists of individual channels' number and its corresponding power.
            ONLY currently ON LED's power will be returned.
        """
        chans = list()
        powers = list()
        self._set_power_unit(unit, irr_distance_mm)
        resp = self._com_query('SCP')
        
        if resp=="":
            self.logger.warning("All LED channels are currently OFF!")
            
        while(resp != ""):
            chans.append(resp.split(',')[0])
            powers.append(resp.split(',')[1])
            resp = self._com.readline()
            
        self.logger.debug(f"All ON LEDs power info received.")
        return (chans,powers)

    #checked
    def get_color_temp(self) -> float:
        """
        - Return the Correlated Color Temperature of the current 
        output spectrum in degrees Kelvin.

        Returns:
        --------
        Color_temp : `float`
            Correlated Color Temperature in degree Kelvin.
        """
        temp = self._com_query('CCT')
        if temp != "":
            try:
                temp=float(temp)
            except Exception:
                temp = -1
        else:
            temp=0
        self.logger.debug(f"Current output spectrum's equivalence color temperature: {temp:.1f} Kelvins.")
        return temp

    #checked
    def get_E_rms_fitted_spectrum(self) -> float:
        """
        - Get Root-Mean-Square error for the fitted specturm vs. provided spectrum.
        
        Returns
        --------
        RMS_Error : `float`
            Root-Mean-Square error for the fitted specturm vs. provided spectrum.
        """
        try:
            erms = float(self._com_query("RPE"))
        except Exception:
            erms = -1
        self.logger.debug(f"Output spectrum matching RMS Error: {erms:.3f}.")
        return erms

    #checked
    def get_chromaticity_output(self) -> 'tuple[float, float]':
        """
        - Return the CIE 1931 chromaticity of the current output spectrum.

        Returns:
        ----------
        (A_CIEx, A_CIEy) : `[float, float]`
            Actual fitted CIEx,y chromaticity in CIE 1931 standard.
        """
        CIExy = list(map(float, self._com_query("OXY").split(',')))
        self.logger.debug(f"Actual output spectrum in CIExy chromaticity {CIExy[0]},{CIExy[1]}.")
        return (CIExy[0], CIExy[1])

    #checked
    def get_tristimulus_output(self) -> 'tuple[float, float, float]':
        """
        - Return the tristimulus of the current output spectrum OSP using 
        theStandard Observer Angle as set by `set_standard_observer_angle`. 
        
        - Note that for 2 degree observer, the tristimulus Y value will 
        represent the output power level in photometric units (lm/m2, cd/m2), 
        identical to what the 'get_output_power_all()' function would return 
        when POWER_UNIT = LUMINANCE or ILLUMINANCE.

        Returns:
        ----------
        (A_CIEx, A_CIEy, A_CIEz) : `[float, float]`
            Actual fitted CIEx,y chromaticity in CIE 1931 standard.
        """
        CIExyz = list(map(float, self._com_query("OXYZ").split(',')))
        self.logger.debug(f"Actual output spectrum in CIExyz space: {CIExyz[0]:.3f},{CIExyz[1]:.3f},{CIExyz[2]:.3f}")
        return (CIExyz[0], CIExyz[1], CIExyz[2])

    #checked
    def get_spectrum_output(self, power_unit:POWER_UNIT=POWER_UNIT.RADIANCE) ->'tuple[list[float], list[float]]':
        """
        - Get the fitted spectrum from the light source, with 1nm step size, and unit
        of radiance or irradiance ONLY. Default range from 360nm to 1100nm i.e. 741 points.

        Parameters
        ----------
        power_unit : 'RS_7_1.POWER_UNIT'
            (defalut = POWER_UNIT.RADIANCE)
        
        Returns
        --------
        spectrum : `tuple[list[float], list[float]]`
            (Wavelength, Specturm data in specified unit with 1nm step size)
        """
        if (power_unit is not self.POWER_UNIT.IRRADIANCE) and (power_unit is not self.POWER_UNIT.RADIANCE):
            self.logger.error("Only radiance and irradiance are supported for spectrum setting!")
            raise bsl_type.DeviceOperationError
        self._set_power_unit(power_unit)
        spectrum = self._com_query("OSP")
        spectrum = list(map(float, spectrum.split(',')))
        self.logger.debug(f"Actual output spectrum received with power unit {power_unit}.")
        lambdas = np.linspace(360,1100,741)
        return (lambdas, spectrum)

    #checked
    def get_spectrum_led(self, led_chan:int, power_unit:POWER_UNIT=POWER_UNIT.RADIANCE) -> 'list[float]':
        """
        - Get the specified LED's realtime power spectrum from the light source, 
        with 1nm step size, and unit of radiance or irradiance ONLY. Default range 
        from 360nm to 1100nm i.e. 741 points.

        Uses
        ----------
        >>> (led_chans,lambda,fwhm) = find_closest_chan(wavelength=650)
        >>> get_led_spectrum(led_chan[0])


        Parameters
        ----------
        led_chan : `int`
            LED channel number as listed in RS_7_1.LED_CHANNELS.LED_CHANS

        power_unit : 'RS_7_1.POWER_UNIT'
            (defalut = POWER_UNIT.RADIANCE)
        
        Returns
        --------
        spectrum : `list[float]`
            Specturm data in specified unit with 1nm step size.
        """
        if (power_unit is not self.POWER_UNIT.IRRADIANCE) and (power_unit is not self.POWER_UNIT.RADIANCE):
            self.logger.error("Only radiance and irradiance are supported for spectrum setting!")
            raise bsl_type.DeviceOperationError
        self._set_power_unit(power_unit)
        spectrum = self._com_query(f"OSP{led_chan}")
        spectrum = list(map(float, spectrum.split(',')))
        self.logger.debug(f"Actual spectrum of LED channel {led_chan} in unit {power_unit} received.")
        return spectrum

    def get_distinct_led_channel_id(self) -> 'tuple(list[int], list[float])':
        """
        Return unique LED channel IDs with nonzero center wavelengths.

        Returns
        -------
        tuple[list[int], list[float]]
            Distinct ``(channel_ids, wavelengths_nm)`` pairs.
        """
        led_chan_id = []
        wavelengths = []
        ids = np.asarray(self.LED_CHANNELS.LEN_CHANS_NO_WHITE.value)
        lambdas = np.asarray(self.LED_CHANNELS.WAVELENGTH.value)
        for id in ids:
            if (lambdas[id] not in wavelengths) and (lambdas[id] != 0):
                led_chan_id.append(id)
                wavelengths.append(lambdas[id])
        return (led_chan_id, wavelengths)

    def close(self) -> None:
        """
        Close RS-7-1 communication resources safely.
        """
        com_obj = getattr(self, "_com", None)
        if com_obj is not None:
            try:
                self.set_power_all(0)
            except Exception:
                pass
            try:
                self.set_iris_position(100)
            except Exception:
                pass
            try:
                com_obj.close()
            except Exception:
                pass
            self._com = None
        self.logger.success(f"CLOSED - Tunable Light Source.\n\n\n")
        pass

    """
    <sorrow-blue>
    Just put something here to make it to the holy glory
    line of 1000th yeah.
    </sorrow-blue>
    """



from pathlib import Path
import numpy as np
from .mantis_file_GS import mantis_file_GS
from numpy.typing import NDArray
from tqdm import tqdm
from loguru import logger
import re
import os

class mantis_folder_GS:
    def __init__(
        self, 
        path: Path, 
        sort_with_exp: bool = False, 
        imager_type: str = "FSI", 
        is_2x2: bool = False, 
        origin=(1,0), 
        R_loc=(0,1), 
        G_loc=(1,0), 
        B_loc=(0,0), 
        SP_loc=(1,1),
        dark_path=None,
        enable_dark_sub=None,    # If None, we decide automatically
        use_filename_exp=True,
        filename_exp_reg=None,
        force_dark_files=True
    ):
        """
        Load a folder of MantisCam .h5 files, optionally do dark subtraction.

        If dark_path is provided but enable_dark_sub is None,
        we force enable_dark_sub = True automatically.
        """
        if isinstance(path, str):
            path = Path(path)
        self.path = path
        self.__videos = dict()

        # If dark_path is provided but user didn't explicitly set enable_dark_sub
        if dark_path is not None and enable_dark_sub is None:
            enable_dark_sub = True

        self.dark_path = dark_path
        self.enable_dark_sub = enable_dark_sub or False  # final bool
        self.use_filename_exp = use_filename_exp
        self.filename_exp_reg = filename_exp_reg
        self.force_dark_files = force_dark_files

        # Attempt to load dark frames if requested
        if self.dark_path and self.enable_dark_sub:
            self._dark_frames = self._load_dark_frames(
                dark_path=self.dark_path, 
                use_filename_exp=self.use_filename_exp, 
                filename_exp_reg=self.filename_exp_reg,
                force_dark_files=self.force_dark_files
            )
        else:
            self._dark_frames = None

        self.__init_mantis_video_dict(
            sort_with_exp=sort_with_exp, 
            imager_type=imager_type, 
            is_2x2=is_2x2, 
            origin=origin, 
            R_loc=R_loc, 
            G_loc=G_loc, 
            B_loc=B_loc, 
            SP_loc=SP_loc
        )

    def init_HDR_for_all(self, tone_mapping="None", mid_tone=0.5, contrast=10, power=0.5):
        """
        Initialize HDR parameters for ALL mantis_file_GS objects in this folder.
        
        Parameters
        ----------
        tone_mapping : {"None", "compress", "enhance"}
            - "None": No tone mapping. 
            - "compress": Simple gamma-based midrange compression.
            - "enhance": Simple sigmoid-based midrange enhancement.
        mid_tone : float
            Mid-tone reference for the "enhance" option.
        contrast : float
            Contrast factor for the "enhance" option.
        power : float
            Gamma exponent if "compress" is selected.
        """
        from loguru import logger
        logger.info(f"Initializing HDR for all files in folder: tone_mapping={tone_mapping}, mid_tone={mid_tone}, contrast={contrast}, power={power}.")
        for video_obj in self.__videos.values():
            video_obj.init_HDR(tone_mapping, mid_tone, contrast, power)

    def __extract_time_key(self, file_name):
        """
        Legacy approach: look for <float>ms in filename for sorting. 
        If not found, fallback to something large so it sorts last.
        """
        match = re.search(r'\d+\.\d+(?=ms)', file_name)
        return float(match.group()) if match else 999999

    def __init_mantis_video_dict(
        self, 
        sort_with_exp: bool = True, 
        imager_type: str = "FSI", 
        is_2x2: bool = False, 
        origin=(1,0), 
        R_loc=(0,1), 
        G_loc=(1,0), 
        B_loc=(0,0), 
        SP_loc=(1,1)
    ):
        filepaths = [
            f for f in self.path.iterdir() 
            if f.is_file() and f.suffix == '.h5'
        ]
        for filepath in filepaths:
            self.__videos[filepath.name] = mantis_file_GS(
                path=filepath,
                imager_type=imager_type,
                is_2x2=is_2x2,
                origin=origin,
                R_loc=R_loc,
                G_loc=G_loc,
                B_loc=B_loc,
                SP_loc=SP_loc,
                dark_frames=self._dark_frames,       # pass the dark dictionary
                enable_dark_sub=self.enable_dark_sub,
                use_filename_exp=self.use_filename_exp,
                filename_exp_reg=self.filename_exp_reg
            )

        if sort_with_exp:
            self.__videos = dict(sorted(
                self.__videos.items(), 
                key=lambda item: self.__extract_time_key(item[0])
            ))

    def __getitem__(self, name: str) -> mantis_file_GS:
        return self.__videos[name]

    @property
    def n_videos(self):
        return len(self.__videos)

    @property
    def name_videos(self):
        filenames = [
            f.name for f in self.path.iterdir() 
            if f.is_file() and f.suffix == '.h5'
        ]
        sorted_filenames = sorted(
            filenames, 
            key=lambda f: os.stat(os.path.join(self.path, f)).st_mtime
        )
        return sorted_filenames
    
    @property
    def arr_files(self) -> list[mantis_file_GS]:
        """
        Return all the file objects in the mantisFolder object.
        """
        return np.array(list(self.__videos.values()))
    
    @property
    def arr_videos(self) -> NDArray[np.uint16]:
        """
        Return all frames in all of the files as a single numpy array.
        [n_files, n_frames, H, W, 1] or something similar.
        Warning: can be huge in memory.
        """
        file_shape = self.arr_files[0].raw_data_shape
        arr_size = (
            self.n_videos * file_shape[0] * file_shape[1] 
            * file_shape[2] * 2 / 1e9
        )
        logger.warning(
            f"Loading entire folder into memory. "
            f"Est. RAM usage = {arr_size:.2f} GB. "
            f"Proceed with caution!"
        )
        return np.array([
            file.frames for file in tqdm(
                self.__videos.values(), 
                desc="Loading Dataset"
            )
        ])
    
    def find_key(self, key: str) -> str:
        """
        Find the key of a file in the folder by substring match.
        """
        for file_name in self.name_videos:
            if key in file_name:
                return file_name
        raise ValueError(f"Key {key} not found in folder {self.path}")

    # --------------------------------------------------------------------------
    # LOAD DARK FRAMES
    # --------------------------------------------------------------------------
    def _load_dark_frames(self, dark_path, use_filename_exp, filename_exp_reg, force_dark_files):
        """
        Load dark frames from either a single .h5 file or all .h5 in a folder,
        returning {exposure_time: 2D mean dark frame}.
        """
        dp = Path(dark_path)
        dark_dict = {}

        if dp.is_file():
            logger.info(f"Loading single dark file: {dp}")
            new_dict = self._load_dark_file(
                dp, use_filename_exp, filename_exp_reg
            )
            dark_dict.update(new_dict)
        else:
            # It's a directory
            all_h5 = [f for f in dp.iterdir() if f.is_file() and f.suffix == '.h5']
            if not force_dark_files:
                all_h5 = [
                    f for f in all_h5 if re.search("dark", f.name, re.IGNORECASE)
                ]
            for f in all_h5:
                logger.info(f"Loading dark file: {f.name}")
                new_dict = self._load_dark_file(
                    f, use_filename_exp, filename_exp_reg
                )
                for expt, dframe in new_dict.items():
                    dark_dict.setdefault(expt, []).append(dframe)

        # Average any duplicates
        for expt in list(dark_dict.keys()):
            frames_list = dark_dict[expt]
            if isinstance(frames_list, list):
                stack = np.stack(frames_list, axis=0)
                dark_dict[expt] = np.mean(stack, axis=0)
        
        logger.info(f"Dark dictionary loaded with keys: {list(dark_dict.keys())}")
        return dark_dict

    def _load_dark_file(self, dark_file: Path, use_filename_exp, filename_exp_reg):
        """
        Return { exptime_float : 2D darkFrame } from this single file,
        using mean across frames[1:]. If only 1 frame, use frames[0:].
        """
        import h5py
        import numpy as np

        with h5py.File(dark_file, 'r') as f:
            frames_data = np.array(f['camera']['frames'][:])  # shape [N, H, W, 1]
            if frames_data.shape[0] <= 1:
                logger.warning(
                    f"Dark file {dark_file.name} has <2 frames. Using all frames for average."
                )
                to_avg = frames_data
            else:
                to_avg = frames_data[1:]
            dark_2d = np.mean(to_avg[..., 0], axis=0)  # [H, W]

            if use_filename_exp:
                exptime = self._parse_exposure_from_filename(dark_file.name, filename_exp_reg)
            else:
                expt_arr = np.array(f['camera']['integration-time-expected'])
                exptime = float(np.median(expt_arr)) if len(expt_arr) else 0.0

        return { exptime: dark_2d }

    def _parse_exposure_from_filename(self, filename: str, filename_exp_reg: str|None):
        """
        If user gave a custom regex, use it. Otherwise:
          1) look for float preceding 'ms'
          2) else last number
          3) else 0.0
        """
        if filename_exp_reg:
            match = re.search(filename_exp_reg, filename)
            if match:
                return float(match.group())
            else:
                logger.warning(
                    f"filename_exp_reg={filename_exp_reg} not found in {filename}, fallback=0.0"
                )
                return 0.0
        match_ms = re.search(r'(\d+\.\d+)(?=ms)', filename)
        if match_ms:
            return float(match_ms.group(1))
        match_num = re.findall(r'(\d+\.?\d*)', filename)
        if match_num:
            return float(match_num[-1])
        logger.warning(f"Unable to parse exptime from {filename}, fallback=0.0")
        return 0.0


from __future__ import annotations

"""
Instrument metadata registry for compatibility-safe object construction.

This layer intentionally stores import paths only. Driver modules are loaded
on demand to avoid changing low-level hardware behavior until construction.
"""

from dataclasses import dataclass


@dataclass(frozen=True)
class InstrumentSpec:
    """
    Static class import metadata for an instrument driver.

    Parameters
    ----------
    key : str
        Logical instrument key.
    module_path : str
        Absolute module import path.
    class_name : str
        Class symbol name inside ``module_path``.
    """

    key: str
    module_path: str
    class_name: str


INSTRUMENT_SPECS = {
    "PM100D": InstrumentSpec(
        key="PM100D",
        module_path="bsl_universal.instruments._inst_lib.instruments._PM100D",
        class_name="PM100D",
    ),
    "BSC203_HDR50": InstrumentSpec(
        key="BSC203_HDR50",
        module_path="bsl_universal.instruments._inst_lib.instruments._BSC203",
        class_name="BSC203_HDR50",
    ),
    "PM400": InstrumentSpec(
        key="PM400",
        module_path="bsl_universal.instruments._inst_lib.instruments._PM400",
        class_name="PM400",
    ),
    "DC2200": InstrumentSpec(
        key="DC2200",
        module_path="bsl_universal.instruments._inst_lib.instruments._DC2200",
        class_name="DC2200",
    ),
    "M69920": InstrumentSpec(
        key="M69920",
        module_path="bsl_universal.instruments._inst_lib.instruments._M69920",
        class_name="M69920",
    ),
    "CS260B": InstrumentSpec(
        key="CS260B",
        module_path="bsl_universal.instruments._inst_lib.instruments._CS260B",
        class_name="CS260B",
    ),
    "HR4000CG": InstrumentSpec(
        key="HR4000CG",
        module_path="bsl_universal.instruments._inst_lib.instruments._HR4000CG",
        class_name="HR4000CG",
    ),
    "RS_7_1": InstrumentSpec(
        key="RS_7_1",
        module_path="bsl_universal.instruments._inst_lib.instruments._RS_7_1",
        class_name="RS_7_1",
    ),
    "SP_2150": InstrumentSpec(
        key="SP_2150",
        module_path="bsl_universal.instruments._inst_lib.instruments._SP_2150",
        class_name="SP_2150",
    ),
    "USB_520": InstrumentSpec(
        key="USB_520",
        module_path="bsl_universal.instruments._inst_lib.instruments._Futek_USB_520",
        class_name="USB_520",
    ),
    "mantisCam": InstrumentSpec(
        key="mantisCam",
        module_path="bsl_universal.instruments._inst_lib.instruments._mantisCam",
        class_name="MantisCamCtrl",
    ),
}

INSTRUMENT_ALIASES = {
    "USB520": "USB_520",
    "MantisCam": "mantisCam",
}


def resolve_instrument_key(name: str) -> str:
    """
    Resolve aliases and validate an instrument key.

    Parameters
    ----------
    name : str
        User-provided key or alias.

    Returns
    -------
    str
        Canonical instrument key.

    Raises
    ------
    KeyError
        If the key is unknown.
    """
    key = INSTRUMENT_ALIASES.get(name, name)
    if key in INSTRUMENT_SPECS:
        return key
    available = ", ".join(sorted(INSTRUMENT_SPECS))
    raise KeyError(f'Unknown instrument "{name}". Available: {available}')


def list_instrument_keys() -> tuple[str, ...]:
    """
    Return all supported canonical instrument keys.

    Returns
    -------
    tuple[str, ...]
        Sorted tuple of supported instrument keys.
    """
    return tuple(sorted(INSTRUMENT_SPECS))



from __future__ import annotations

"""
Robust serial interface wrapper with retry and reconnect behavior.
"""

import platform
import re
import subprocess
import time
from typing import Any, Callable

from loguru import logger

try:
    import serial
    from serial.tools.list_ports import comports
except Exception:
    serial = None

    def comports():
        return []

from ..headers._bsl_inst_info import _bsl_inst_info_list
from ..headers._bsl_type import _bsl_type as bsl_type

try:
    from bsl_universal.core.device_health import device_health_hub
except Exception:
    device_health_hub = None

logger_opt = logger.opt(ansi=True)


class _bsl_serial:
    """
    Serial transport helper used by instrument drivers.

    Parameters
    ----------
    target_inst : _bsl_inst_info_list
        Instrument metadata descriptor.
    device_sn : str, optional
        Optional serial selector for device matching, by default ``""``.
    """

    MAX_CONNECT_RETRY = 3
    MAX_IO_RETRY = 3
    RETRY_DELAY_SEC = 0.2

    def __init__(self, target_inst: _bsl_inst_info_list, device_sn: str = "") -> None:
        """
        Initialize serial transport helper.

        Parameters
        ----------
        target_inst : _bsl_inst_info_list
            Instrument metadata descriptor.
        device_sn : str, optional
            Optional serial selector for device matching, by default ``""``.
        """
        logger_opt.info("    Initiating bsl_serial_service...")
        self.device_id = ""
        self.inst = target_inst
        self.target_device_sn = device_sn
        self.serial_port_name = None
        self.baudrate = self.inst.BAUDRATE if self.inst.BAUDRATE else 9600
        if serial is None:
            self.serial_port = None
            logger_opt.error(
                "pyserial is not installed; serial communication unavailable for {}.",
                self.inst.MODEL,
            )
            return
        self.serial_port = self._connect_serial_device()
        if self.serial_port is None:
            logger_opt.error(
                "<light-blue><italic>{} ({})</italic></light-blue> not found on serial ports.",
                self.inst.MODEL,
                self.target_device_sn,
            )

    def __del__(self) -> None:
        self.close()
        return None

    def _connect_serial_device(self) -> serial.Serial | None:
        """
        Connect to a matching serial device with bounded retries.

        Returns
        -------
        serial.Serial | None
            Open serial object when successful, otherwise ``None``.
        """
        if serial is None:
            return None

        for attempt in range(1, self.MAX_CONNECT_RETRY + 1):
            if self._find_device():
                try:
                    ser = serial.Serial(self.serial_port_name, self.baudrate, timeout=0.1)
                    logger_opt.success(
                        "    {} with DEVICE_ID: <light-blue><italic>{}</italic></light-blue> found and connected!",
                        self.inst.MODEL,
                        self.device_id,
                    )
                    return ser
                except Exception as exc:
                    logger_opt.warning(
                        "Open serial port failed on attempt {}/{}: {}",
                        attempt,
                        self.MAX_CONNECT_RETRY,
                        exc,
                    )
            time.sleep(self.RETRY_DELAY_SEC)
        return None

    def _reconnect(self) -> bool:
        """
        Attempt to restore serial connection.

        Returns
        -------
        bool
            True when reconnection succeeds.
        """
        if serial is None:
            return False

        try:
            if self.serial_port is not None:
                self.serial_port.close()
        except Exception:
            pass
        self.serial_port = None

        # Fast path: reopen known port and baud first.
        if self.serial_port_name:
            try:
                self.serial_port = serial.Serial(
                    self.serial_port_name,
                    self.baudrate,
                    timeout=0.1,
                )
                return True
            except Exception:
                self.serial_port = None

        # Full rescan fallback.
        self.serial_port = self._connect_serial_device()
        return self.serial_port is not None

    def _find_device(self) -> bool:
        """
        Search COM ports for a matching instrument.

        Returns
        -------
        bool
            True if a candidate device is found and validated.
        """
        if serial is None:
            return False

        com_ports_list = list(comports())
        logger_opt.trace("    Devices found on bus: {}", [p[0] for p in com_ports_list])

        for port in com_ports_list:
            temp_port = None

            if self.inst.SERIAL_SN in port[0]:
                logger_opt.info(
                    "    Specified device <light-blue><italic>{}</italic></light-blue> with Serial SN <light-blue><italic>{}</italic></light-blue> found on port <light-blue><italic>{}</italic></light-blue> by Device Serial SN search.",
                    self.inst.MODEL,
                    self.inst.SERIAL_SN,
                    port[0],
                )
                temp_port = port[0]

            if self.target_device_sn and self.target_device_sn in port[0]:
                self.device_id = port[0]
                logger_opt.info(
                    "    Specified device <light-blue><italic>{}</italic></light-blue> with Serial SN <light-blue><italic>{}</italic></light-blue> found on port <light-blue><italic>{}</italic></light-blue> by Device Serial SN search.",
                    self.inst.MODEL,
                    self.target_device_sn,
                    port[0],
                )
                temp_port = port[0]

            if self.inst.SERIAL_NAME in port[1]:
                logger_opt.info(
                    "    Specified device <light-blue><italic>{}</italic></light-blue> with Serial_Name <light-blue><italic>{}</italic></light-blue> found on port <light-blue><italic>{}</italic></light-blue> by Device name search.",
                    self.inst.MODEL,
                    self.inst.SERIAL_NAME,
                    port[0],
                )
                temp_port = port[0]

            if (self.inst.USB_PID in port[2]) or (self._hex_to_dec(self.inst.USB_PID) in port[2]):
                logger_opt.info(
                    "    Specified device <light-blue><italic>{}</italic></light-blue> with USB_PID: <light-blue><italic>{}</italic></light-blue> found on port <light-blue><italic>{}</italic></light-blue> by USB_PID search.",
                    self.inst.MODEL,
                    self.inst.USB_PID,
                    port[0],
                )
                temp_port = port[0]

            if temp_port is not None:
                found_port, baudrate = self._check_device_resp(temp_port)
                if found_port is not None:
                    self.baudrate = baudrate
                    self.serial_port_name = found_port
                    return True

        if self.target_device_sn and self.inst.MODEL == "USB_520":
            logger_opt.error(
                "<light-blue><italic>{} ({})</italic></light-blue> not found on serial ports.",
                self.inst.MODEL,
                self.target_device_sn,
            )
            raise bsl_type.DeviceConnectionFailed

        logger.warning("    No device found based on USB_PID/VID or Serial Name search!")

        # Full brute-force probe fallback.
        for port in com_ports_list:
            found_port, baudrate = self._check_device_resp(port[0])
            if found_port is not None:
                self.baudrate = baudrate
                self.serial_port_name = found_port
                return True
        return False

    def _hex_to_dec(self, value: str) -> str:
        """
        Convert hex string to decimal string.

        Parameters
        ----------
        value : str
            Hexadecimal string.

        Returns
        -------
        str
            Decimal representation or empty string on failure.
        """
        try:
            return str(int(value, 16))
        except Exception:
            return ""

    def _check_device_resp(self, temp_port: str) -> tuple[str | None, int | None]:
        """
        Probe a serial port for expected identification response.

        Parameters
        ----------
        temp_port : str
            Candidate serial port path.

        Returns
        -------
        tuple[str | None, int | None]
            Matched port and baudrate on success, otherwise ``(None, None)``.
        """
        if serial is None:
            return None, None

        if self.inst.BAUDRATE != 0:
            baudrates = [self.inst.BAUDRATE]
        else:
            baudrates = [4800, 9600, 19200, 28800, 38400, 115200]

        if not self.is_port_free(temp_port):
            logger_opt.warning(
                "    BUSY - Device <light-blue><italic>{}</italic></light-blue> is busy, moving to next available device...",
                temp_port,
            )
            return None, None

        try:
            for baudrate in baudrates:
                logger_opt.info(
                    "    Inquiring serial port <light-blue><italic>{}</italic></light-blue> with Baudrate={}",
                    temp_port,
                    baudrate,
                )
                with serial.Serial(temp_port, baudrate, timeout=0.1) as device:
                    logger_opt.trace(
                        "        Connected to <light-blue><italic>{}</italic></light-blue> on port <light-blue><italic>{}</italic></light-blue>",
                        device.name,
                        temp_port,
                    )
                    if self.inst.QUERY_CMD != "N/A":
                        device.reset_input_buffer()
                        device.write(bytes(self.inst.QUERY_CMD, "utf-8"))
                        logger_opt.trace(
                            "        Query <light-blue><italic>{}</italic></light-blue> sent to <light-blue><italic>{}</italic></light-blue>",
                            repr(self.inst.QUERY_CMD),
                            device.name,
                        )
                        time.sleep(0.5)
                    try:
                        resp = repr(device.read(100).decode("utf-8")).strip("\n\r")
                    except Exception:
                        resp = "ERROR in interpreting as UTF-8."
                    logger_opt.trace(
                        "        Response from <light-blue><italic>{}</italic></light-blue>: {}",
                        device.name,
                        resp,
                    )

                    if self.inst.QUERY_E_RESP in resp:
                        if self.inst.QUERY_SN_CMD == "N/A":
                            return temp_port, baudrate

                        logger_opt.info(
                            "        <light-blue><italic>{}</italic></light-blue> found on serial bus on port <light-blue><italic>{}</italic></light-blue>.",
                            self.inst.MODEL,
                            temp_port,
                        )
                        device.reset_input_buffer()
                        device.write(bytes(self.inst.QUERY_SN_CMD, "utf-8"))
                        time.sleep(0.5)
                        resp = device.read(100).decode("utf-8").strip("\n\r")
                        logger_opt.trace(
                            "        Response from <light-blue><italic>{}</italic></light-blue>: {}",
                            device.name,
                            resp,
                        )
                        match = re.search(self.inst.SN_REG, resp)
                        device_id = match.group(0) if match is not None else "UNKNOWN"
                        if self.target_device_sn in device_id or not self.target_device_sn:
                            self.device_id = device_id.strip("\r\n")
                            return temp_port, baudrate
                        logger_opt.warning(
                            "    S/N Mismatch - Device <light-blue><italic>{}</italic></light-blue> with S/N <light-blue><italic>{}</italic></light-blue> found, not <light-blue><italic>{}</italic></light-blue> as requested, moving to next available device...",
                            temp_port,
                            device_id,
                            self.target_device_sn,
                        )
                        break
        except Exception:
            logger_opt.warning(
                "    BUSY - Device <light-blue><italic>{}</italic></light-blue> is busy, moving to next available device...",
                temp_port,
            )
            return None, None
        return None, None

    def _report_unrecoverable_failure(self, message: str) -> None:
        """
        Report unrecoverable transport failure to runtime health registry.

        Parameters
        ----------
        message : str
            Error message.
        """
        if device_health_hub is None:
            return
        try:
            device_health_hub.register_failure(
                instrument_key=str(self.inst.MODEL),
                model=str(self.inst.MODEL),
                device_type=str(getattr(self.inst, "TYPE", "Unknown")),
                serial_number=str(self.device_id or self.target_device_sn or "Unknown"),
                error_message=message,
                unrecoverable=True,
            )
        except Exception:
            pass

    def _run_with_reconnect(
        self,
        operation_name: str,
        callback: Callable[[], Any],
    ) -> Any:
        """
        Execute an operation with reconnect retries.

        Parameters
        ----------
        operation_name : str
            Label used for logs and failure reports.
        callback : Callable[[], Any]
            Operation callback.

        Returns
        -------
        Any
            Callback return value.

        Raises
        ------
        bsl_type.DeviceOperationError
            If retries are exhausted.
        """
        serial_errors = (OSError, AttributeError, UnicodeDecodeError)
        if serial is not None:
            serial_errors = (serial.SerialException,) + serial_errors

        for attempt in range(1, self.MAX_IO_RETRY + 1):
            if self.serial_port is None and not self._reconnect():
                time.sleep(self.RETRY_DELAY_SEC)
                continue
            try:
                return callback()
            except serial_errors as exc:
                logger_opt.warning(
                    "Serial {} attempt {}/{} failed for {}: {}",
                    operation_name,
                    attempt,
                    self.MAX_IO_RETRY,
                    self.inst.MODEL,
                    exc,
                )
                self._reconnect()
                time.sleep(self.RETRY_DELAY_SEC)

        message = (
            f"{self.inst.MODEL} serial {operation_name} failed after "
            f"{self.MAX_IO_RETRY} attempts."
        )
        logger_opt.error(message)
        self._report_unrecoverable_failure(message)
        raise bsl_type.DeviceOperationError(message)

    def readline(self) -> str:
        """
        Read one line from the serial stream.

        Returns
        -------
        str
            UTF-8 decoded line with trailing CR/LF removed.
        """
        def _run() -> str:
            resp = self.serial_port.readline().decode("utf-8")
            logger_opt.trace(
                "        {} - com-Serial - Resp from {} with {}",
                self.inst.MODEL,
                self.inst.MODEL,
                repr(resp),
            )
            return resp.strip("\n\r")

        return self._run_with_reconnect("readline", _run)

    def read(self, n_bytes: int) -> str:
        """
        Read fixed number of bytes from the serial stream.

        Parameters
        ----------
        n_bytes : int
            Number of bytes to read.

        Returns
        -------
        str
            UTF-8 decoded payload with trailing CR/LF removed.
        """
        def _run() -> str:
            resp = self.serial_port.read(n_bytes).decode("utf-8")
            logger_opt.trace(
                "        {} - com-Serial - Resp from {} with {}",
                self.inst.MODEL,
                self.inst.MODEL,
                repr(resp),
            )
            return resp.strip("\n\r")

        return self._run_with_reconnect("read", _run)

    def read_all(self) -> bytes:
        """
        Read all currently buffered bytes.

        Returns
        -------
        bytes
            Buffered payload with surrounding whitespace removed.
        """
        def _run() -> bytes:
            resp = self.serial_port.read_all()
            logger_opt.trace(
                "        {} - com-Serial - Resp from {} with {}",
                self.inst.MODEL,
                self.inst.MODEL,
                repr(resp),
            )
            return resp.strip()

        return self._run_with_reconnect("read-all", _run)

    def write(self, msg: str) -> int:
        """
        Write bytes to serial stream.

        Parameters
        ----------
        msg : str
            Payload string to write.

        Returns
        -------
        int
            Number of bytes written.
        """
        def _run() -> int:
            logger_opt.trace(
                "        {} - com-Serial - Write to {} with {}",
                self.inst.MODEL,
                self.inst.MODEL,
                repr(msg),
            )
            return self.serial_port.write(bytes(msg, "utf-8"))

        return self._run_with_reconnect("write", _run)

    def writeline(self, msg: str) -> int:
        """
        Write a CRLF-terminated line to serial stream.

        Parameters
        ----------
        msg : str
            Line payload (without CRLF).

        Returns
        -------
        int
            Number of bytes written.
        """
        def _run() -> int:
            line = msg + "\r\n"
            logger_opt.trace(
                "        {} - com-Serial - Write to {} with {}",
                self.inst.MODEL,
                self.inst.MODEL,
                repr(line),
            )
            return self.serial_port.write(bytes(line, "utf-8"))

        return self._run_with_reconnect("writeline", _run)

    def query(self, cmd: str) -> str:
        """
        Execute a write/read query cycle.

        Parameters
        ----------
        cmd : str
            Command payload written as a line.

        Returns
        -------
        str
            Device response line.
        """
        def _run() -> str:
            self.serial_port.reset_input_buffer()
            line = cmd + "\r\n"
            self.serial_port.write(bytes(line, "utf-8"))
            resp = self.serial_port.readline().decode("utf-8")
            return resp.strip("\n\r")

        return self._run_with_reconnect("query", _run)

    def flush_read_buffer(self) -> None:
        """
        Clear incoming serial buffer.
        """
        def _run() -> None:
            self.serial_port.reset_input_buffer()

        self._run_with_reconnect("flush", _run)

    def set_serial_timeout(self, timeout: int) -> None:
        """
        Set serial read timeout in seconds.

        Parameters
        ----------
        timeout : int
            Timeout in seconds.
        """
        def _run() -> None:
            self.serial_port.timeout = timeout

        self._run_with_reconnect("set-timeout", _run)

    def is_port_free(self, port_name: str) -> bool:
        """
        Check whether a serial port is available on macOS.

        Parameters
        ----------
        port_name : str
            Full path to the serial device, e.g. ``/dev/tty.usbserial``.

        Returns
        -------
        bool
            True when the port appears free.
        """
        if platform.system() != "Darwin":
            return True

        try:
            result = subprocess.run(
                ["lsof", port_name],
                capture_output=True,
                text=True,
                check=False,
            )
            if port_name in result.stdout:
                return False
            return True
        except Exception as exc:
            logger_opt.warning("Port availability check failed for {}: {}", port_name, exc)
            return True

    def close(self) -> None:
        """
        Close the serial port safely.
        """
        try:
            if self.serial_port is not None:
                self.serial_port.close()
        except Exception:
            pass
        finally:
            self.serial_port = None



from .factory import InstrumentFactory, available_instruments, create_instrument, factory
from .inst import (
    BSC203_HDR50,
    CS260B,
    DC2200,
    HR4000CG,
    M69920,
    PM100D,
    PM400,
    RS_7_1,
    SP_2150,
    USB_520,
    init_logger,
    mantisCam,
)
from ..core.instrument_runtime import ManagedInstrument, RecoveryPolicy, SafeInstrumentProxy

__all__ = [
    "init_logger",
    "PM100D",
    "BSC203_HDR50",
    "PM400",
    "DC2200",
    "M69920",
    "CS260B",
    "HR4000CG",
    "RS_7_1",
    "SP_2150",
    "USB_520",
    "mantisCam",
    "InstrumentFactory",
    "factory",
    "create_instrument",
    "available_instruments",
    "ManagedInstrument",
    "RecoveryPolicy",
    "SafeInstrumentProxy",
]



from ._thorlabs_apt_device.devices import BSC
from ..headers._bsl_inst_info import _bsl_inst_info_list as inst
from ..headers._bsl_logger import _bsl_logger as bsl_logger
from ..headers._bsl_type import _bsl_type as bsl_type
import time
import numpy as np

class BSC203(BSC):
    """
    A class for ThorLabs APT device model BSC203.

    It is based off :class:`BSC`, but looking for a serial number starting with ``"70"`` and setting ``x = 1`` (should be ``x = 3``)--FOR FUTURE.

    :param serial_port: Serial port device the device is connected to.
    :param vid: Numerical USB vendor ID to match.
    :param pid: Numerical USB product ID to match.
    :param manufacturer: Regular expression to match to a device manufacturer string.
    :param product: Regular expression to match to a device product string.
    :param serial_number: Regular expression to match to a device serial number.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse" directions.
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    """
    def __init__(self, serial_port=None, vid=None, pid=None, manufacturer=None, product=None, serial_number="70", location=None, home=False, invert_direction_logic=False, swap_limit_switches=True):
        """
        Initialize base BSC203 stage controller.

        Parameters
        ----------
        serial_port : str | None, optional
            Explicit serial port path, by default None.
        vid : int | None, optional
            USB vendor ID matcher, by default None.
        pid : int | None, optional
            USB product ID matcher, by default None.
        manufacturer : str | None, optional
            Manufacturer matcher, by default None.
        product : str | None, optional
            Product matcher, by default None.
        serial_number : str, optional
            Device serial matcher, by default ``"70"``.
        location : str | None, optional
            USB location matcher, by default None.
        home : bool, optional
            Home on startup when True, by default False.
        invert_direction_logic : bool, optional
            Invert forward/reverse polarity, by default False.
        swap_limit_switches : bool, optional
            Swap limit switch mapping, by default True.
        """
        super().__init__(
            serial_port=serial_port,
            vid=vid,
            pid=pid,
            manufacturer=manufacturer,
            product=product,
            serial_number=serial_number,
            location=location,
            x=1,
            home=home,
            invert_direction_logic=invert_direction_logic,
            swap_limit_switches=swap_limit_switches,
        )

class BSC203_HDR50(BSC203):
    """
    A class for ThorLabs APT device model BSC203 with the HDR50 rotating stage.

    It is based off :class:`BSC203`, but with sensible homing and velocity parameters.

    For the HDR50, there are 66*409600 microsteps per revolution of the stage, defined in the datasheet

    :param closed_loop: Boolean to indicate the use of an encoded stage (the "E" in LNR502E) in closed-loop mode, default is False.
    :param serial_port: Serial port device the device is connected to.
    :param vid: Numerical USB vendor ID to match.
    :param pid: Numerical USB product ID to match.
    :param manufacturer: Regular expression to match to a device manufacturer string.
    :param product: Regular expression to match to a device product string.
    :param serial_number: Regular expression to match to a device serial number.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse" directions.
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    """
    
   
    MOTOR_STAGE_RATIO = 66
    MOTOR_FREV_STEP = 409600
    MOTOR_RUN_STEP = int(0.45 / 360 * MOTOR_STAGE_RATIO * MOTOR_FREV_STEP)


    def __init__(self, serial_port=None, vid=None, pid=None, manufacturer=None, product=None, serial_number="70", location=None, home=False, invert_direction_logic=False, swap_limit_switches=True):
        """
        Initialize HDR50 rotating-stage controller configuration.

        Parameters
        ----------
        serial_port : str | None, optional
            Explicit serial port path, by default None.
        vid : int | None, optional
            USB vendor ID matcher, by default None.
        pid : int | None, optional
            USB product ID matcher, by default None.
        manufacturer : str | None, optional
            Manufacturer matcher, by default None.
        product : str | None, optional
            Product matcher, by default None.
        serial_number : str, optional
            Device serial matcher, by default ``"70"``.
        location : str | None, optional
            USB location matcher, by default None.
        home : bool, optional
            Home on startup when True, by default False.
        invert_direction_logic : bool, optional
            Invert forward/reverse polarity, by default False.
        swap_limit_switches : bool, optional
            Swap limit switch mapping, by default True.
        """
        self._init_kwargs = dict(
            serial_port=serial_port,
            vid=vid,
            pid=pid,
            manufacturer=manufacturer,
            product=product,
            serial_number=serial_number,
            location=location,
            home=home,
            invert_direction_logic=invert_direction_logic,
            swap_limit_switches=swap_limit_switches,
        )
        self.inst = inst.BSC203_HDR50
        self.logger = bsl_logger(self.inst)
        self. __curr_step = None
        super().__init__(serial_port, vid, pid, manufacturer, product, serial_number, location, home, invert_direction_logic, swap_limit_switches)
        for bay_i, _ in enumerate(self.bays):
            for channel_i, _ in enumerate(self.channels):
                self.set_velocity_params(acceleration=self.MOTOR_FREV_STEP//10, max_velocity=100*self.MOTOR_FREV_STEP, bay=bay_i, channel=channel_i)
                #velocity params
                self.set_home_params(velocity=100*self.MOTOR_FREV_STEP,offset_distance=224000, bay=bay_i, channel=channel_i) 
                #homing params, might need adjustment individually
        self.logger.warning("Due to OPEN LOOP control, HDR50 is prone to errors, regular homing required.")
        time.sleep(0.3)
        self._home()
        self.logger.success("initial homing complete")
        return


    def __del__(self, *args, **kwargs) -> None:
        try:
            self.close()
        except Exception:
            pass
        return None

    def reconnect(self, rehome: bool = True) -> bool:
        """
        Recover stage control and optionally perform homing.

        Parameters
        ----------
        rehome : bool, optional
            Perform homing after recovery, by default True.

        Returns
        -------
        bool
            True when recovery succeeds.
        """
        try:
            if rehome:
                self.home()
            else:
                self._blocker()
            return True
        except Exception as exc:
            self.logger.error(f"Failed to recover BSC203 HDR50: {type(exc)}")
            return False

    def reset_stage(self, rehome: bool = True) -> bool:
        """
        Reset stage motion state by re-homing when requested.

        Parameters
        ----------
        rehome : bool, optional
            Re-home stage to zero reference, by default True.

        Returns
        -------
        bool
            True when reset succeeds.
        """
        return self.reconnect(rehome=rehome)
    
    def _home(self, bay=0, channel=0):
        """
        Cause the rotation stage to rotate to its mechanical “home” position. 
        This should result in the marking to be pointing at "0"
        This is a blocking function
        """
        super().home(bay, channel)
        self.__curr_step = 0
        return self._blocker()
    
    def step(self, stepnum, called_by_home = False):
        """
        move in steps of 0.45 degrees, with 800 steps a revolution (home before use is advised)
        no negative values allowed for steps--FOR FUTURE UPGRADES
        This is a blocking function
        """
        if stepnum == 0:
            return self._blocker()

        if stepnum < 0:
            stepnum = 800-abs(stepnum)
            self.logger.warning("NEGATIVE VALUES NOT ALLOWED! GOING TO DESTINATION in POSITIVE direction...")

        if called_by_home:
            super().move_relative(stepnum*self.MOTOR_RUN_STEP)
            return self._blocker()

        if stepnum <= 100:    
            super().move_relative(stepnum*self.MOTOR_RUN_STEP)
        else:
            self.set_velocity_params(acceleration=self.MOTOR_FREV_STEP//10, max_velocity=200*self.MOTOR_FREV_STEP, bay=0, channel=0)
            time.sleep(0.3)
            super().move_relative(stepnum*self.MOTOR_RUN_STEP)
            self.set_velocity_params(acceleration=self.MOTOR_FREV_STEP//10, max_velocity=100*self.MOTOR_FREV_STEP, bay=0, channel=0)
            time.sleep(0.3)
        
        
        
        try:
            self.__curr_step += stepnum
            self.__curr_step = self.__curr_step % 800
        except:
            self.home()
            self.logger.error("did not home on startup, homing...")
        
        return self._blocker()
    

    def get_curr_angle(self, return_angle = False):
        """
        get the current OPEN LOOP angle of the device, 
        this angle will be close to the actual value but there is no guarentees.
        This is a none-blocking function, do not SPAM
        """
        if self.__curr_step == None: 
            self.logger.error(f"FAILED to get current angle on the HDR50 rotation stage, please home and try again!\n\n\n")
            return None
        self.logger.info(f"Current angle at {self.__curr_step * 0.45} degrees, or {self.__curr_step} steps.")
        if return_angle:
            return self.__curr_step * 0.45
        else:
            return 

    def _is_moving(self):
        """
        True: the system is moving
        False: the system is not moving
        """
        time.sleep(0.3)
        return (self.status['moving_forward'] or self.status['moving_reverse'])
    
    def _blocker(self):
        """
        private function to make functions blocking, when the motor is busy, wait.
        """
        time.sleep(0.3)
        while self._is_moving():
            time.sleep(0.1)
        return 

    def home(self, bay=0, channel=0):
        """
        Optimized homing algorithm. If the current position is very close to 0, it will home normally. 
        But if the current position is far from home, it will home in crazy mode. 
        Be careful, DON'T touch the stage when you use this--Bill Yang
        """
        stepnum = 800 - self.__curr_step
        if stepnum >= 750:
            self._home()
            self.__curr_step = 0
        else:
            self.set_velocity_params(acceleration=self.MOTOR_FREV_STEP//10, max_velocity=1000*self.MOTOR_FREV_STEP, bay=0, channel=0)
            time.sleep(0.3)
            self.step(stepnum-10, called_by_home=True)
            self.set_velocity_params(acceleration=self.MOTOR_FREV_STEP//10, max_velocity=100*self.MOTOR_FREV_STEP, bay=0, channel=0)
            time.sleep(0.3)
            self._home(bay, channel)
            self.__curr_step = 0
        return self._blocker()
    
    def set_angle(self, angle:int, precision = False):
        """
        set the angle of the HDR50
        precision == False; mode does not home before moving
        precision == True; mode homes before moving

        """
        doable_step = int(np.floor(angle/0.45))
        doable_angle = doable_step * 0.45
        if doable_angle != angle:
            self.logger.warning(f"angle given not divisible by step size, moving to step {doable_step}, angle {doable_angle}")
            self.logger.warning(f"angle given not divisible by step size, your error is {angle - doable_angle}")
        else:
            self.logger.info(f"moving to step {doable_step}, angle {doable_angle}")
        if precision == True:    
            self.home()
            self.step(doable_step)
            self.logger.success(f"absolute angle manuver complete, step at {doable_step}, angle at {doable_angle}")
        else:
            self.step(doable_step-self.__curr_step)
        return self._blocker()

    def close(self) -> None:
        """
        Close BSC203 controller safely.
        """
        try:
            super().close()
        except Exception:
            pass



from __future__ import annotations

"""
Compatibility-safe instrument factory.

This sits above legacy drivers and does not alter underlying command/transport
flows, enabling architecture refactoring with minimal hardware risk.
"""

from importlib import import_module
import time
from typing import Any

from .registry import INSTRUMENT_SPECS, list_instrument_keys, resolve_instrument_key


class InstrumentFactory:
    """
    Lazy-loading instrument factory with retry safety.

    Notes
    -----
    The factory retries constructor calls to reduce transient initialization
    failures on hardware buses and converts ``SystemExit`` into typed exceptions.
    """

    def __init__(self):
        """
        Initialize an empty instrument class cache.
        """
        self._class_cache: dict[str, type] = {}

    def _get_class(self, name: str) -> type:
        """
        Resolve and cache the instrument class for ``name``.

        Parameters
        ----------
        name : str
            Logical instrument key.

        Returns
        -------
        type
            Instrument class object.
        """
        key = resolve_instrument_key(name)
        if key in self._class_cache:
            return self._class_cache[key]
        spec = INSTRUMENT_SPECS[key]
        module = import_module(spec.module_path)
        cls = getattr(module, spec.class_name)
        self._class_cache[key] = cls
        return cls

    def create(
        self,
        name: str,
        *args: Any,
        retries: int = 2,
        retry_delay_sec: float = 0.5,
        **kwargs: Any,
    ) -> Any:
        """
        Construct an instrument instance with retry protection.

        Parameters
        ----------
        name : str
            Logical instrument key.
        *args : Any
            Positional args for the instrument constructor.
        retries : int, optional
            Number of constructor retries after the first attempt, by default 2.
        retry_delay_sec : float, optional
            Delay between retries in seconds, by default 0.5.
        **kwargs : Any
            Keyword args for the instrument constructor.

        Returns
        -------
        Any
            Constructed instrument object.

        Raises
        ------
        RuntimeError
            If construction fails after retries.
        Exception
            Re-raises the last constructor exception when available.
        """
        cls = self._get_class(name)
        attempts = max(1, int(retries) + 1)
        last_exc: Exception | None = None

        for attempt in range(1, attempts + 1):
            try:
                return cls(*args, **kwargs)
            except SystemExit as exc:
                last_exc = RuntimeError(
                    f'Instrument "{name}" initialization attempted to terminate process '
                    f"(SystemExit: {exc})."
                )
            except Exception as exc:
                last_exc = exc

            if attempt < attempts:
                time.sleep(max(0.0, float(retry_delay_sec)))

        if last_exc is not None:
            raise last_exc
        raise RuntimeError(f'Failed to initialize instrument "{name}".')

    def available(self) -> tuple[str, ...]:
        """
        List available logical instrument keys.

        Returns
        -------
        tuple[str, ...]
            Sorted available keys.
        """
        return list_instrument_keys()


factory = InstrumentFactory()


def create_instrument(name: str, *args: Any, **kwargs: Any) -> Any:
    """
    Convenience wrapper for ``InstrumentFactory.create``.

    Parameters
    ----------
    name : str
        Logical instrument key.
    *args : Any
        Positional constructor args.
    **kwargs : Any
        Keyword constructor args.

    Returns
    -------
    Any
        Constructed instrument object.
    """
    return factory.create(name, *args, **kwargs)


def available_instruments() -> tuple[str, ...]:
    """
    Convenience wrapper for ``InstrumentFactory.available``.

    Returns
    -------
    tuple[str, ...]
        Sorted available keys.
    """
    return factory.available()



"""
Central logger configuration for bsl_universal.
"""

import sys

from loguru import logger

_STDOUT_FORMAT = (
    "<cyan>{time:MM-DD at HH:mm:ss}</cyan> | <level>{level:7}</level> | "
    "{file:15}:{line:4} | <level>{message}</level>"
)
_STDERR_FORMAT = (
    "<green>{time:HH:mm:ss}</green> | <level>{level}</level> | "
    "{function}:{line} - <level>{message}</level>"
)


def init_logger(log_level: str = "DEBUG") -> None:
    """
    Initialize process-level logging for library usage.

    Parameters
    ----------
    log_level : str, optional
        Loguru log level string, by default ``"DEBUG"``.
    """
    logger.remove()
    logger.add(
        sys.stdout,
        colorize=True,
        format=_STDOUT_FORMAT,
        level=log_level,
        diagnose=False,
    )
    logger.add(
        sys.stderr,
        colorize=True,
        format=_STDERR_FORMAT,
        level=log_level,
        diagnose=False,
    )
    logger.success(f'Logger initialized with LOG_LEVEL = "{log_level}".')



# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["BBD201", "BBD202", "BBD203"]

from .. import protocol as apt
from .aptdevice_motor import APTDevice_BayUnit
from ..enums import EndPoint


class BBD201(APTDevice_BayUnit):
    """
    A class for ThorLabs APT device model BBD201.

    It is based off :class:`APTDevice_BayUnit` configured with a single channel and the the BBD serial number
    preset in the initialisation method.

    :param serial_port: Serial port device the device is connected to.
    :param vid: Numerical USB vendor ID to match.
    :param pid: Numerical USB product ID to match.
    :param manufacturer: Regular expression to match to a device manufacturer string.
    :param product: Regular expression to match to a device product string.
    :param serial_number: Regular expression matching the serial number of device to search for.
    :param location: Regular expression to match to a device bus location.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse" directions.
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    """
    def __init__(self, serial_port=None, vid=None, pid=None, manufacturer=None, product=None, serial_number="73", location=None, home=True, invert_direction_logic=False, swap_limit_switches=True):

        super().__init__(serial_port=serial_port, vid=vid, pid=pid, manufacturer=manufacturer, product=product, serial_number=serial_number, location=location, x=1, home=home, invert_direction_logic=invert_direction_logic, swap_limit_switches=swap_limit_switches, status_updates="auto")


class BBD202(APTDevice_BayUnit):
    """
    A class for ThorLabs APT device model BBD202.

    It is based off :class:`BBD_BSC` configured with two channels and the the BBD serial number
    preset in the initialisation method.

    :param serial_port: Serial port device the device is connected to.
    :param vid: Numerical USB vendor ID to match.
    :param pid: Numerical USB product ID to match.
    :param manufacturer: Regular expression to match to a device manufacturer string.
    :param product: Regular expression to match to a device product string.
    :param serial_number: Regular expression to match to a device serial number.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse" directions.
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    """
    def __init__(self, serial_port=None, vid=None, pid=None, manufacturer=None, product=None, serial_number="73", location=None, home=True, invert_direction_logic=False, swap_limit_switches=True):

        super().__init__(serial_port=serial_port, vid=vid, pid=pid, manufacturer=manufacturer, product=product, serial_number=serial_number, location=location, x=2, home=home, invert_direction_logic=invert_direction_logic, swap_limit_switches=swap_limit_switches, status_updates="auto")


class BBD203(APTDevice_BayUnit):
    """
    A class for ThorLabs APT device model BBD203.

    It is based off :class:`BBD_BSC` configured with three channels and the the BBD serial number
    preset in the initialisation method.

    :param serial_port: Serial port device the device is connected to.
    :param vid: Numerical USB vendor ID to match.
    :param pid: Numerical USB product ID to match.
    :param manufacturer: Regular expression to match to a device manufacturer string.
    :param product: Regular expression to match to a device product string.
    :param serial_number: Regular expression to match to a device serial number.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse" directions.
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    """
    def __init__(self, serial_port=None, vid=None, pid=None, manufacturer=None, product=None, serial_number="73", location=None, home=True, invert_direction_logic=False, swap_limit_switches=True):

        super().__init__(serial_port=serial_port, vid=vid, pid=pid, manufacturer=manufacturer, product=product, serial_number=serial_number, location=location, x=3, home=home, invert_direction_logic=invert_direction_logic, swap_limit_switches=swap_limit_switches, status_updates="auto")




# Copyright 2021 Patrick C. Tapping
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["BSC", "BSC201", "BSC201_DRV250"]

from .. import protocol as apt
from .aptdevice_motor import APTDevice_BayUnit
from ..enums import EndPoint

class BSC(APTDevice_BayUnit):
    """
    A class for the BSC10x and BSC20x series of Thorlabs APT controllers.

    It is based off
    :class:`~thorlabs_apt_device.devices.aptdevice_motor.APTDevice_BayUnit` with the
    addition of the control loop parameters used for encoded stages.

    Note that it seems (at least some of) the BSC series will send automatic status updates, but
    they do not support the corresponding acknowledgement message, and so will stop responding after
    ~5 seconds due to the lack of acknowledgement from the host. As a workaround, the status of the
    device will be polled.

    Additionally, it seems that the initial movement and homing velocities can be effectively zero,
    making it seem like the device is not working, though it is actually just moving extremely
    slowly. It's a good idea to ensure sensible values are set for
    :meth:`~thorlabs_apt_device.devices.aptdevice_motor.APTDevice_Motor.set_velocity_params`,
    :meth:`~thorlabs_apt_device.devices.aptdevice_motor.APTDevice_Motor.set_jog_params` and
    :meth:`~thorlabs_apt_device.devices.aptdevice_motor.APTDevice_Motor.set_home_params` during
    initialisation. What values qualify as "sensible" will depend on the particular stepper and
    stage being controlled. For example, the DRV250 actuator has 409600 microsteps/mm, 21987328
    microsteps/mm/s, and 4506 microsteps/mm/s/s. Thus, using ``acceleration=4506`` and
    ``max_velocity=21987328`` in
    :meth:`~thorlabs_apt_device.devices.aptdevice_motor.APTDevice_Motor.set_velocity_params` would
    result in a movement speed of 1 mm/s and an acceleration of 1 mm/s/s. Consult the `APT
    Communications Protocol
    <https://www.thorlabs.com/Software/Motion%20Control/APT_Communications_Protocol.pdf>`__ PDF
    document on the Thorlabs website for parameters for a specific stage.

    :param serial_port: Serial port device the device is connected to.
    :param vid: Numerical USB vendor ID to match.
    :param pid: Numerical USB product ID to match.
    :param manufacturer: Regular expression to match to a device manufacturer string.
    :param product: Regular expression to match to a device product string.
    :param serial_number: Regular expression matching the serial number of device to search for.
    :param location: Regular expression to match to a device bus location.
    :param x: Number of channels the device controls (the x in BSC20x).
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse" directions.
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    """
    def __init__(self, serial_port=None, vid=None, pid=None, manufacturer=None, product=None, serial_number=None, location=None, home=True, x=1, invert_direction_logic=False, swap_limit_switches=True):

        super().__init__(serial_port=serial_port, vid=vid, pid=pid, manufacturer=manufacturer, product=product, serial_number=serial_number, location=location, home=home, x=x, invert_direction_logic=invert_direction_logic, swap_limit_switches=swap_limit_switches, status_updates="polled")

        self.loopparams_ = [[{
            "loop_mode": 0,
            "prop": 0,
            "int": 0,
            "diff": 0,
            "pid_clip": 0,
            "pid_tol": 0,
            "encoder_const": 0,
            # Update message fields
            "msg" : "",
            "msgid" : 0,
            "source" : 0,
            "dest" : 0,
            "chan_ident" : 0,
        } for _ in self.channels] for _ in self.bays]
        """
        Parameters for the closed-loop position control for encoded stages.

        Fields are ``"loop_mode"``, ``"prop"``, ``"int"``, ``"diff"``, ``"pid_clip"``, ``"pid_tol"``, and ``"encoder_const"``.
        """
        # Request current control loop parameters
        for bay in self.bays:
            for channel in self.channels:
                self._loop.call_soon_threadsafe(self._write, apt.mot_req_kcubekstloopparams(source=EndPoint.HOST, dest=bay, chan_ident=channel))

        self.powerparams_ = [[{
            "rest_factor": 0,
            "move_factor": 0,
            # Update message fields
            "msg" : "",
            "msgid" : 0,
            "source" : 0,
            "dest" : 0,
            "chan_ident" : 0,
        } for _ in self.channels] for _ in self.bays]
        """
        Parameters for the motor power parameters.

        Fields are ``"rest_factor"`` and ``"move_factor"``.
        """
        # Request current motor power parameters
        for bay in self.bays:
            for channel in self.channels:
                self._loop.call_soon_threadsafe(self._write, apt.mot_req_powerparams(source=EndPoint.HOST, dest=bay, chan_ident=channel))

        if x == 1:
            self.loopparams = self.loopparams_[0][0]
            """Alias to first bay/channel of :data:`loopparams_`."""
            self.powerparams = self.powerparams_[0][0]
            """Alias to first bay/channel of :data:`powerparams`."""

    def _process_message(self, m):
        super()._process_message(m)

        # Decode bay and channel IDs and check if they match one of ours
        if m.msg in ("mot_get_kcubekstloopparams", "mot_get_powerparams"):
            # Check if source matches one of our bays
            try:
                bay_i = self.bays.index(m.source)
            except ValueError:
                # Ignore message from unknown bay id
                self._log.warn(f"Message {m.msg} has unrecognised source={m.source}.")
                bay_i = 0
            # Check if channel matches one of our channels
            try:
                channel_i = self.channels.index(m.chan_ident)
            except ValueError:
                # Ignore message from unknown channel id
                self._log.warn(f"Message {m.msg} has unrecognised channel={m.chan_ident}.")
                channel_i = 0
        
        if m.msg == "mot_get_kcubekstloopparams":
            # Control loop update message
            self.loopparams_[bay_i][channel_i].update(m._asdict())
        elif m.msg == "mot_get_powerparams":
            # Motor power update message
            self.powerparams_[bay_i][channel_i].update(m._asdict())


    def set_loop_params(self, loop_mode=None, prop=None, integral=None, diff=None, pid_clip=None, pid_tol=None, encoder_const=None, bay=0, channel=0):
        """
        Configure the closed-loop positioning parameters used by encoded stages.

        The closed-loop control algorithm is a PID style. Note that the keyword parameter to modify
        the "int" value is named "integral" so as not to confuse it with the python int keyword.

        :param loop_mode: Set to 1 if open-loop, 2 for closed-loop mode.
        :param prop: Coefficient of the proportional gain term.
        :param integral: Coefficient of the integral gain term.
        :param diff: Coefficient of the differential gain term.
        :param pid_clip: Maximum value of the PID algorithm output.
        :param pid_tol: Minimum allowed (non-zero) value of the PID algorithm output.
        :encoder_const: Coefficient to map encoder counts to motor steps.
        """
        if loop_mode is None:
            loop_mode = self.loopparams_[bay][channel]["loop_mode"]
        if prop is None:
            prop = self.loopparams_[bay][channel]["prop"]
        if integral is None:
            integral = self.loopparams_[bay][channel]["int"]
        if diff is None:
            diff = self.loopparams_[bay][channel]["diff"]
        if pid_clip is None:
            pid_clip = self.loopparams_[bay][channel]["pid_clip"]
        if pid_tol is None:
            pid_tol = self.loopparams_[bay][channel]["pid_tol"]
        if encoder_const is None:
            encoder_const = self.loopparams_[bay][channel]["encoder_const"]

        self._log.debug(f"Setting control loop parameters for [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        self._loop.call_soon_threadsafe(self._write, apt.mot_set_kcubekstloopparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], loop_mode=loop_mode, prop=prop, int=integral, diff=diff, pid_clip=pid_clip, pid_tol=pid_tol, encoder_const=encoder_const))
        # Update status with new loop parameters
        self._loop.call_soon_threadsafe(self._write, apt.mot_req_kcubekstloopparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel]))


    def set_power_params(self, rest_factor=None, move_factor=None, bay=0, channel=0):
        """
        Configure the motor power parameters used during movement and rest.

        Values are expressed as a percentage between ``0`` and ``100``. Typically move power should
        be set to ``100``, and the rest power reduced to prevent excessive heating of the motor.


        :param rest_factor: Percentage power used when at rest.
        :param move_factor: Percentage power used when moving.
        """
        if rest_factor is None:
            rest_factor = self.loopparams_[bay][channel]["rest_factor"]
        if move_factor is None:
            move_factor = self.loopparams_[bay][channel]["move_factor"]
        
        self._log.debug(f"Setting motor power parameters for [bay={self.bays[bay]:#x}, channel={self.channels[channel]}].")
        self._loop.call_soon_threadsafe(self._write, apt.mot_set_powerparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel], rest_factor=rest_factor, move_factor=move_factor))
        # Update status with new power parameters
        self._loop.call_soon_threadsafe(self._write, apt.mot_req_powerparams(source=EndPoint.HOST, dest=self.bays[bay], chan_ident=self.channels[channel]))


class BSC201(BSC):
    """
    A class for ThorLabs APT device model BSC201.

    It is based off :class:`BSC`, but looking for a serial number starting with ``"40"`` and setting ``x = 1``.

    :param serial_port: Serial port device the device is connected to.
    :param vid: Numerical USB vendor ID to match.
    :param pid: Numerical USB product ID to match.
    :param manufacturer: Regular expression to match to a device manufacturer string.
    :param product: Regular expression to match to a device product string.
    :param serial_number: Regular expression to match to a device serial number.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse" directions.
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    """
    def __init__(self, serial_port=None, vid=None, pid=None, manufacturer=None, product=None, serial_number="40", location=None, home=True, invert_direction_logic=False, swap_limit_switches=True):

        super().__init__(serial_port=serial_port, vid=vid, pid=pid, manufacturer=manufacturer, product=product, serial_number=serial_number, location=location, x=1, home=home, invert_direction_logic=invert_direction_logic, swap_limit_switches=swap_limit_switches)


class BSC201_DRV250(BSC201):
    """
    A class for ThorLabs APT device model BSC201 with DRV250 stepper-motor-driven actuator, sold as
    a package as the `LNR502 and LNR502E <https://www.thorlabs.com/newgrouppage9.cfm?objectgroup_id=2297&pn=LNR502E/M>`__.

    It is based off :class:`BSC201`, but with sensible default movement parameters configured for the actuator.

    For the DRV250, there are 409600 microsteps/mm, 21987328 microsteps/mm/s, and 4506 microsteps/mm/s/s.

    For the models with the optical encoder (LNR502E, LNR502E/M), use the parameter
    ``closed_loop=True`` to enable closed-loop positioning. This will configure the controller
    parameters to use feedback from the encoder during positioning. For non-encoded stages, or to
    use open-loop control, this should be set to ``closed_loop=False``.

    :param closed_loop: Boolean to indicate the use of an encoded stage (the "E" in LNR502E) in closed-loop mode, default is False.
    :param serial_port: Serial port device the device is connected to.
    :param vid: Numerical USB vendor ID to match.
    :param pid: Numerical USB product ID to match.
    :param manufacturer: Regular expression to match to a device manufacturer string.
    :param product: Regular expression to match to a device product string.
    :param serial_number: Regular expression to match to a device serial number.
    :param home: Perform a homing operation on initialisation.
    :param invert_direction_logic: Invert the meaning of "forward" and "reverse" directions.
    :param swap_limit_switches: Swap "forward" and "reverse" limit switch values.
    """
    def __init__(self, serial_port=None, vid=None, pid=None, manufacturer=None, product=None, serial_number="40", location=None, home=True, invert_direction_logic=False, swap_limit_switches=True, closed_loop=False):

        super().__init__(serial_port=serial_port, vid=vid, pid=pid, manufacturer=manufacturer, product=product, serial_number=serial_number, location=location, home=home, invert_direction_logic=invert_direction_logic, swap_limit_switches=swap_limit_switches)

        # Initial velocity parameters are effectively zero on startup, set something more sensible
        # Homing is initiated 1.0s after init, so hopefully these will take effect before then...
        for bay_i, _ in enumerate(self.bays):
            for channel_i, _ in enumerate(self.channels):
                self.set_velocity_params(acceleration=4506, max_velocity=21987328, bay=bay_i, channel=channel_i)
                self.set_jog_params(size=409600, acceleration=4506, max_velocity=21987328, bay=bay_i, channel=channel_i)
                self.set_home_params(velocity=21987328, offset_distance=20480, bay=bay_i, channel=channel_i)
    
        if closed_loop:
            # Configure some default parameters for the closed-loop positioning routine
            for bay_i, _ in enumerate(self.bays):
                for channel_i, _ in enumerate(self.channels):
                    self.set_loop_params(loop_mode=2, prop=50000, integral=5000, diff=100, pid_clip=16000000, pid_tol=80, encoder_const=4292282941, bay=0, channel=0)
        else:
            # Use open-loop positioning (only using stepper counts)
            for bay_i, _ in enumerate(self.bays):
                for channel_i, _ in enumerate(self.channels):
                    self.set_loop_params(loop_mode=1, prop=0, integral=0, diff=0, pid_clip=0, pid_tol=0, encoder_const=0, bay=0, channel=0)

